// $Id: Mapgen.C,v 1.9 2005/11/06 10:01:06 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

#include "Mapgen.H"
#include "Map.H"
#include "Tile.H"
#include "MapTool.H"

using namespace std;

// initialize RNG depending on cmd line parameter

static void init_rand(Random &rand)
{
  if (MapTool::opt.seed) {
    rand.srand(MapTool::opt.seed);
  } else {
    rand.srand(time(0));
  }
}


void Mapgen::init_map(Map<Tile> &map_data)
{
  FORS (i, map_data.get_num()) {
    map_data(i).set_type(GROUND);
  }
}

void Mapgen::generate_obstacle_map(Map<Tile> &map_data, sint4 min_size, sint4 max_size, sint4 num, real8 d)
{
  assert(min_size <= max_size);
  
  Random rand;
  Point p;  
  Point offset;
  sint4 size;

  init_rand(rand);

  FORS (i, num) {

    offset.x = 0;
    offset.y = 0;
    
    rand.srand(rand.rand_uint4());

//    p.x = rand.rand_uint4() % map_data.get_width();
//    p.y = rand.rand_uint4() % map_data.get_height();

    //center
    p.x = map_data.get_width()/2;
    p.y = map_data.get_height()/2;
        
    //offset
    offset.x = ((rand.rand_uint4()%(map_data.get_width() ) ))- map_data.get_width()/2;
    offset.y = ((rand.rand_uint4()%(map_data.get_height()) ))- map_data.get_height()/2;
    
//    cout << "[1] x " << offset.x << " y " << offset.y <<endl;
    
    if (d) {
      //center favoring handicap

      if (offset.x < 0) { offset.x += rand.rand_uint4()%-offset.x/(1/d); }
      if (offset.x > 0) { offset.x -= rand.rand_uint4()% offset.x/(1/d); }
    
      if (offset.y < 0) { offset.y += rand.rand_uint4()%-offset.y/(1/d); }
      if (offset.y > 0) { offset.y -= rand.rand_uint4()% offset.y/(1/d); }
    }
    
    p.x += offset.x;
    p.y += offset.y;
    
//    cout << "[2] x " << p.x << " y " << p.y <<endl;
    
    size = rand.rand_uint4() % max_size - min_size;
    size += min_size;
    
    MapTool::raise_island(map_data, 4, p, size, rand.rand_uint4());
  }
}


void Mapgen::generate_basecamp_map(Map<Tile> &map_data, stringstream &objects,
				   sint4 num_players, sint4 size, sint4 margin)
{
  Random rand;
  init_rand(rand);
  
  // determine the larges radias (outer + inner)/2
  real8 island_rad = sqrt(((real8) size))/2 + margin;
  real8 outer = (real8)min(map_data.get_width(), map_data.get_height());
  real8 inner = outer - (real8)island_rad;
  real8 rad = (inner/2 + outer/2)/2;
 
  real8 rotation = (rand.rand_uint4() % 360)/(M_PI/180);
  
  real8 cx, cy, angle;
   
  cx = map_data.get_width()/2;
  cy = map_data.get_height()/2;
  angle = 2*M_PI/(real8)num_players;
    
  rand.srand(time(0));
  
  Point point;

  FORS (i, num_players) {
  
    point.x = (int)(cx + cos(angle*(i) + rotation) * rad); 
    point.y = (int)(cy + sin(angle*(i) + rotation) * rad);  

    MapTool::raise_island(map_data, 0, point , size, rand.rand_uint4());
    MapTool::place_unit(objects, "start_loc", i, point.x, point.y);
  }
  
}


/**   
 * starting at the north west corner of the map.
 * if a split occures the north west part of the tile
 * keeps its heights, and south east part is synchronized
 * with the west and south neighbours.
 * */

void Mapgen::finalize_map(Map<Tile> &map_data)
{
  sint4 h0, h1, h2, h3;

  FORS (j, map_data.get_height()) {

    FORS (i, map_data.get_width()) {

      h0 = h1 = h2 = h3 = 0; 
      
      if (i != 0 && j != 0) h0 = map_data(i-1, j-1).get_height(4); //NW SE
      if (          j != 0) h1 = map_data(i  , j-1).get_height(4); //N  SE
                            h2 = map_data(i  , j  ).get_height(4); //.  SE
      if (i != 0          ) h3 = map_data(i-1, j  ).get_height(4); //W  SE

      // 4 and O
      if (h0 == h1 && h1 == h2 && h2 == h3) {
        map_data(i, j).set_topo(h0, h1, h2, h3, Tile::NO_SPLIT);
        map_data(i, j).set_type(GROUND);
        continue;
      }

      if (h0 < 0) h0 = 0;
      if (h1 < 0) h1 = 0;
      if (h2 < 0) h2 = 0;
      if (h3 < 0) h3 = 0;

      if (h0 == h1 && h2 == h3 || h0 == h3 && h1 == h2 || h0 == h2 && h1 == h3) {
        //--------------------
        map_data(i, j).set_topo(h0, h1, h2, h3, Tile::NO_SPLIT);
        map_data(i, j).set_type(CLIFF, CLIFF);
        continue;
      }

      //TB
      if (h0 == h1 && h1 == h2) {
        map_data(i, j).set_topo(h0, h1, h2, h3, Tile::TB_SPLIT);
        map_data(i, j).set_type(CLIFF, GROUND);
        if (h0 < 0) {
          map_data(i, j).set_type(WATER, GROUND);
        }
        continue;
      }

      if (h0 == h3 && h3 == h2) {
        map_data(i, j).set_topo(h0, h1, h2, h3, Tile::TB_SPLIT);
        map_data(i, j).set_type(GROUND, CLIFF);
        if (h0 < 0) {
          map_data(i, j).set_type(GROUND, WATER);
        }
        continue;
      }

      if (h0 == h1 && h1 == h3) {
        map_data(i, j).set_topo(h0, h1, h2, h3, Tile::BT_SPLIT);
        map_data(i, j).set_type(GROUND, CLIFF);
        if (h0 < 0) {
          map_data(i, j).set_type(GROUND, WATER);
        }
        continue;
      }

      if (h1 == h2 && h2 == h3) {
        map_data(i, j).set_topo(h0, h1, h2, h3, Tile::BT_SPLIT);
        map_data(i, j).set_type(CLIFF, GROUND);
        if (h1 < 0) {
          map_data(i, j).set_type(WATER, GROUND);
        }
        continue;
      }

      cout << "\nUNHANDLED:" << endl;
      cout << "\t" << h0 << " " << h1 << endl;
      cout << "\t" << endl;
      cout << "\t" << h3 << " " << h2 << endl;
    }
  }
}
