#ifndef Game1_ST_BasicMove_H
#define Game1_ST_BasicMove_H

#include "Game1_SimpleTerrain.H"

namespace Game1_SimpleTerrain {

  /// The definition of a square using two corners.
  struct UnitSquare {
    Loc topLeft;
    Loc bottomRight;
    const Object *obj;
    ST_Task *task;

    UnitSquare() : obj(NULL), task(NULL) {};
    UnitSquare(Loc topLeft_, Loc bottomRight_, const Object *obj_, ST_Task *task_)
      : topLeft(topLeft_), bottomRight(bottomRight_), obj(obj_), task(task_) {};
    ~UnitSquare() {};

    UnitSquare& operator=(const UnitSquare& other) {
      if( this != &other ) {
        topLeft = other.topLeft;
        bottomRight = other.bottomRight;
        obj = other.obj;
        task = other.task;
      }
      return *this;
    }
    
  };

  /** A sector is a large area, containing objects (actually tasks) whose paths 
      in the near future will pass through the area, and hence may collide. */
  struct Sector {
    Vector<ST_Task*> tasks;
    
    Sector() {};
    Sector(const Sector &other) : tasks(other.tasks) {};
    Sector& operator=(const Sector &other) { 
      if( this != &other ) 
        tasks = other.tasks; 
      return *this; 
    };
  };


  /// Simple Movement-execution module
  class ST_BasicMove {
  public:
    ST_BasicMove(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_);
    ~ST_BasicMove();

    /// add command-line options specific to this implementation, using static methods of class Options
    static void add_options(void);

    /// Monitor task execution, return move commands and status messages.
    void execute_tasks(TerrainBasicImp<ST_Task>::Tasks& tasks, 
                       ST_Terrain::Objects &objsToStop,
                       Vector<MoveCmd> &cmds, 
                       Vector<StatusMsg> &msgs);

    //---------------------------------------------------------------------
    /// \name Changes to the world.
    //---------------------------------------------------------------------
    //@{
    void insert_object(const Object *obj);
    void remove_object(const Object *obj);
    void update_object(const Object *obj);
    void insert_boundary(const Segment& line);
    //@}
    //---------------------------------------------------------------------

  private:

    // prevent copying as we currently don't need it.
    ST_BasicMove(const ST_BasicMove& other);
    ST_BasicMove& operator=(const ST_BasicMove& other);


    //---------------------------------------------------------------------
    /// \name Internal accessories
    //---------------------------------------------------------------------
    //@{
    /** Test if the object is at the given location, 
	accounting for a 1-tick look-ahead for moving objects. */
    static bool is_at_location(const Object* obj, const Loc& target);
    /** Test if the distance from the object to the target is within the given range.*/
    static bool is_near_location(const Object* obj, const Loc& target, sint4 range);
    /** Test if the object is at the goal of the given task, implementing 
        (some) of the various definitions of 'goal'. */
    bool is_at_goal(const Object* obj, const Task& task, const ST_Terrain::TIter &st_task);

#if 0
    /** Test if the object is within firing range */
    static bool is_in_range(const Object* moving_obj, const Object *goal_obj);
    /** Test if the location is within firing range */
    static bool is_in_range(const Object* moving_obj, const Loc goal_loc);
#endif 

    /// Create a MoveCmd for the current waypoint of the given UnitPath, and add to the vector.
    static void add_move_command(const UnitPath& path, Vector<MoveCmd> &cmds);
    //@}
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    /// \name Local obstacle avoidance
    //---------------------------------------------------------------------
    //@{
    /// Anticipate and handle collisions that are expected to occur in the next X ticks.
    static const sint4 ANTICIPATE_COLLISION_TICKS;
    /// Compute expected collisions using integral coordinates but with a higher accuracy.
    static const sint4 FINE_GRID_RES_BITS;

    /// Identify and resolve expected collisions between controlled units.
    void resolve_friendly_collisions(ST_Terrain::Tasks &tasks, ST_Terrain::Objects &objsToStop);
    /** true iff the objects of the given tasks will collide before 
        reaching their next waypoint, and within the ANTICIPATE_COLLISION_TICKS 
        next ticks. */
    bool will_collide(Vector<ST_Task*>::iterator i, 
		      Vector<ST_Task*>::iterator j, 
                      bool assume_obj1_stopped=false, 
                      bool assume_obj2_stopped=false);
    /// Convert to the higher resolution
    static sint4 fine_grid(sint4 v) { return v << FINE_GRID_RES_BITS; };
    /** \brief Get the position of a given object and its heading towards a given target, using fine coordinates.
        Compute the spatial params (position, target, velocity) using the fine-grid
        coordinates.
        @param o              The object
        @param target         The location it is moving to, in world (not fine) coordinates. This will be translated to fine coords as output.
        @param fine_position  If != NULL, will contain o's position in fine coords, as output.
        @param fine_velocity  If != NULL, will contain o's velocity (direction+speed) in fine coords, as output.
    */
    void get_fine_spatial_params(const Object *o, ScalarPoint &target,
                                 ScalarPoint *fine_position, ScalarPoint *fine_heading);
    //@}
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    /// \name Motion Sectors for faster collision detection - James Bergsma
    //---------------------------------------------------------------------
    //@{
    /// How many tiles per motion sector (side)
    static const uint4 TILES_PER_MOTION_SECTOR;

    /** Find the position of an object in 'when' ticks. 
        For multiple-waypoints paths, this is only an approximation:
        1. It assumes a continuous motion, disregarding the rounding of the
           time it takes to travel each straight line leg of the path to 
           integer ticks (rounding done by the ORTS engine).
        2. It doesn't consider delays in giving the "go to next 
           waypoint" actions (eg due to network lag or slow client sise 
           processing), which cause the unit to stop at waypoints.

        @param  cur_path  The path to analyze.
        @param  when      Compute the location in that many ticks.
        @param  where     Returns the future position in this parameter.
        @return the waypoint number (0 is first waypoint, located at the back() 
                of cur_path.path) that will already be passed through by 
                that time, -1 if the first waypoint will not be reached.
    */
    sint4 get_future_position(const UnitPath &cur_path, sint4 when, Loc &where);

    /** Using a future position create a square containing the object. 
        @param task   The task to analyze. Assuming that task->path is valid.
        @return       The UnitSquare that will contain the object in the close future.
    */   
    UnitSquare find_motion_square(const ST_Terrain::TIter &task);

    /** Insert each motion square into the sectors that contain it. Sectors
        with >= 2 squares are marked as active (meaning they should be 
        tested for potential collisions. 
	@param motionSquares    These are the squares in which units are currently active.
	@param sectors          This is a large grid representation in which the sectors will be placed.
    */
    void insert_into_sectors(const Vector<UnitSquare> &motionSquares,
                                         SimpleMap<Sector> &sectors);
    //@}
    //---------------------------------------------------------------------

    Random rand;
    bool noMotionSectors;                 ///< Should the motion sectors optimization be used for collision detection
    sint4 tiles_x, tiles_y, tile_points; 
  };

} // end of namespace SimpleTerrain

REGISTER_TYPEOF(2551, Vector<Game1_SimpleTerrain::Sector>::iterator);
REGISTER_TYPEOF(2552, Vector<Game1_SimpleTerrain::Sector>::const_iterator);
REGISTER_TYPEOF(2553, Vector<Game1_SimpleTerrain::Sector*>::iterator);
REGISTER_TYPEOF(2554, Vector<Game1_SimpleTerrain::Sector*>::const_iterator);
REGISTER_TYPEOF(2555, Vector<Game1_SimpleTerrain::UnitSquare>::iterator);
REGISTER_TYPEOF(2556, Vector<Game1_SimpleTerrain::UnitSquare>::const_iterator);

#endif
