#include "GLWidget.H"

/**
 * File glfractal.cpp
 * Brief Implementation of GLWidget
 * This is a simple QGLWidget displaying an openGL wireframe box
 *
 * The OpenGL code is mostly borrowed from Brian Pauls "spin" example
 * in the Mesa distribution.
 * $ID$
 */

/**
 * Constructor that creates a GLWidget widget
 */
GLWidget::GLWidget(QWidget* parent, const std::string& name)
  : QGLWidget(parent, name.c_str())
{
   xRot = yRot = zRot = 0.0;
   scale = 1.25;
   object = 0;

   mapTileWidth = mapTileHeight= 64;
   mouseXPosition = mouseYPosition = -1;
   leftMouseDown = middleMouseDown = rightMouseDown = false;
}

GLWidget::~GLWidget()
{  glDeleteLists( object, 1 );
}

/*--------------------------------------------------------------------------*/
/**
 * Paint the box. The actual openGL commands for drawing the box are
 * performed here.
 */
void GLWidget::paintGL()
{
  using namespace std;

  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  if(!Worlded::isInitialized())
    return;

  glLoadIdentity();
/*
  //glEnable( GL_DEPTH_TEST ); 
  glTranslatef(0.0, 0.0, -10.0);
  // set the zoom according to the scale variable
  glScalef(scale, scale, scale);
  // set the image rotation up according to xRot, yRot, zRot
  glRotatef( xRot, 1.0, 0.0, 0.0);
  glRotatef( yRot, 0.0, 1.0, 0.0);
  glRotatef( zRot, 0.0, 0.0, 1.0);
*/
  //glCallList(object);
  //drawGrid();

  Worlded *world = Worlded::instance();

  Map<Tile> *worldMap = (Worlded::instance())->get_map();
  
  GLfloat xInc = GLfloat(viewportWidth)  / worldMap->get_width();
  GLfloat yInc = GLfloat(viewportHeight) / worldMap->get_height();

  MyPoint curTopLeftCorner;
  curTopLeftCorner.x = -xInc;    // Set these this way so the first iteration
  curTopLeftCorner.y = -yInc;    // through the loop sets this to (0,0)

  // Draw the tiles in the map.
  for(int i = 0 ;  i < worldMap->get_height() ; i++){
    curTopLeftCorner.y += yInc;
    curTopLeftCorner.x = -xInc;
    for(int j = 0 ; j < worldMap->get_width() ; j++){
      curTopLeftCorner.x += xInc;
      drawTile(worldMap->get_tile(i,j), curTopLeftCorner, xInc, yInc);
    }
  }

  drawGrid();
}


void GLWidget::drawTile(const Tile& t, const MyPoint& tlc, GLfloat width, GLfloat height){

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  if(t.is_half()){

    GLfloat halfWidth = width / 2;

    glBegin(GL_QUADS);

      // West Half Tile
      setColor(t.get_typeW(), t.get_height(0));
      glVertex2d(tlc.x, tlc.y);
      glVertex2d(tlc.x, tlc.y + height);
      glVertex2d(tlc.x + halfWidth, tlc.y + height);
      glVertex2d(tlc.x + halfWidth, tlc.y);

      // East Half Tile
      setColor(t.get_typeE(), t.get_height(0));
      glVertex2d(tlc.x + halfWidth, tlc.y);
      glVertex2d(tlc.x + halfWidth, tlc.y + height);
      glVertex2d(tlc.x + width, tlc.y + height);
      glVertex2d(tlc.x + width, tlc.y);

    glEnd();
    
  }
  else if(t.is_split()){

    if(Tile::TB_SPLIT == t.get_split()){

      glBegin(GL_TRIANGLES);

        setColor(t.get_typeW(), t.get_height(0));
        glVertex2d(tlc.x, tlc.y);
        glVertex2d(tlc.x, tlc.y + height);
        glVertex2d(tlc.x + width, tlc.y + height);
      
        setColor(t.get_typeE(), t.get_height(0));
        glVertex2d(tlc.x + width, tlc.y);
        glVertex2d(tlc.x , tlc.y);
        glVertex2d(tlc.x + width, tlc.y + height);

      glEnd();
    }
    else{ // (BT_SPLIT)

      glBegin(GL_TRIANGLES);

        setColor(t.get_typeW(), t.get_height(0));
        glVertex2d(tlc.x, tlc.y);
        glVertex2d(tlc.x, tlc.y + height);
        glVertex2d(tlc.x + width, tlc.y);
      
        setColor(t.get_typeE(), t.get_height(0));
        glVertex2d(tlc.x + width, tlc.y);
        glVertex2d(tlc.x , tlc.y + height);
        glVertex2d(tlc.x + width, tlc.y + height);

      glEnd();
    }
  }
  else{ // Draw the tile ..

    glBegin(GL_QUADS);

      setColor(t.get_typeW(), t.get_height(0));
      glVertex2d(tlc.x, tlc.y);
      glVertex2d(tlc.x, tlc.y + height);
      glVertex2d(tlc.x + width, tlc.y + height);
      glVertex2d(tlc.x + width, tlc.y);

    glEnd();
    

  }

  // Draw tile outline
  /*
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  
  glColor3f(0.0,0.0,0.0);
  
  glBegin(GL_QUADS);
    glVertex2d(tlc.x, tlc.y);
    glVertex2d(tlc.x, tlc.y + height);
    glVertex2d(tlc.x + width, tlc.y + height);
    glVertex2d(tlc.x + width, tlc.y);
  glEnd();
  */

}

void GLWidget::drawGrid(){

  glColor3f(0.0, 0.0, 0.0);

  // Draw all the horizontal grid lines
  GLfloat curX = 0.0;
  GLfloat xInc = GLfloat(viewportWidth) / ((Worlded::instance())->get_map())->get_width();
  for( ; curX <= viewportWidth ; curX += xInc){
    glBegin(GL_LINES);
      glVertex2f(curX, 0);
      glVertex2f(curX, viewportHeight);
    glEnd();
  }

  // Draw all the vertical grid lines
  GLfloat curY = 0.0;
  GLfloat yInc = GLfloat(viewportHeight) / ((Worlded::instance())->get_map())->get_height();
  for( ; curY <= viewportHeight ; curY += yInc){
    glBegin(GL_LINES);
      glVertex2f(0, curY);
      glVertex2f(viewportWidth, curY);
    glEnd();
  }
}


void GLWidget::setColor(uint4 tileType, sint4 height){
  
  if(tileType == gGROUND)
    glColor3f(0.0, 0.5 + (float(15) / height) / 2, 0.0);
  else if(tileType == gCLIFF)
    glColor3f(1.0, 1.0, 0.0);
  else // default to water....
    glColor3f(0.0, 0.0, 1.0);
}

/*--------------------------------------------------------------------------*/
/**
 * Set up the OpenGL rendering state, and define display list
 */

void GLWidget::initializeGL()
{
  glClearColor( 1.0, 1.0, 1.0, 1.0 ); // Let OpenGL clear to black
  glShadeModel( GL_FLAT ); // we want smooth shading . . . try GL_FLAT if you like
}

/*--------------------------------------------------------------------------*/
/**
 * Set up the OpenGL view port, matrix mode, etc.
 */
void GLWidget::resizeGL( int w, int h )
{
  viewportWidth  = w;
  viewportHeight = h;

  glViewport( 0, 0, (GLint)w, (GLint)h );

  glMatrixMode( GL_PROJECTION );

  glLoadIdentity();

  gluOrtho2D(-5, w+5, -5, h+5); 

  glMatrixMode( GL_MODELVIEW );
  updateGL();
}

//*********************************************************
// Functions to rotate the model along the x, y, & z axes.
void GLWidget::setXRotation(int degrees)
{  xRot=(GLfloat)(degrees % 360);
   updateGL();
}

void GLWidget::setYRotation(int degrees)
{  yRot=(GLfloat)(degrees % 360);
   updateGL();
}

void GLWidget::setZRotation(int degrees)
{  zRot=(GLfloat)(degrees % 360);
   updateGL();
}

//**********************************************************
// Write Your Own Function To Set The Scale Here!!!!


void GLWidget::mousePressEvent(QMouseEvent *e){

  updateGL();

  /*
  if(e->button() == LeftButton)
    leftMouseDown = true;
  else if(e->button() == RightButton)
    rightMouseDown = true;
  else if(e->button() == MidButton)
    middleMouseDown = true;

  mouseXPosition = e->x();
  mouseYPosition = e->y();
  */
}

void GLWidget::mouseReleaseEvent(QMouseEvent *e){
  /*
  if(e->button() == LeftButton)
    leftMouseDown = false;
  else if(e->button() == RightButton)
    rightMouseDown = false;
  else if(e->button() == MidButton)
    middleMouseDown = false;
  */
}

void GLWidget::mouseMoveEvent(QMouseEvent *e){
  /*
  if(leftMouseDown){

    yRot += mouseXPosition - e->x();
    xRot += mouseYPosition - e->y();

    mouseXPosition = e->x();
    mouseYPosition = e->y();

    updateGL();
  }
  */
}


