#ifndef GfxAnimesh_H
#define GfxAnimesh_H

// $Id: GfxAnimesh.H,v 1.52 2007/03/21 17:16:02 orts_mburo Exp $

// This is an ORTS file (c) Keith Yerex, licensed under the GPL

#include "GfxGlobal.H"
#include "GfxObject.H"
#include "Shader.H"

class GfxModule;

//===================================================================

struct Poly
{
  sint2 vert[3];
};

struct ModelTag
{
  Mat4<real4> axis;
};

struct BBox
{
  Vec3<real4> min;
  Vec3<real4> max;
};

//-------------------------------------------------------------------

/** A model with keyframed animations. */

class Model
{
friend class GfxAnimesh;
public:

  sint4 load_textures(const std::string &filename, real4 gamma);

  virtual sint4 load(const std::string &modelfile, const std::string &texfile,
		     real4 scale, real4 dy, real4 gamma) = 0;
  
  virtual void draw_bbox(sint4 frame) = 0;
  virtual void draw(sint4 frame) = 0;
  virtual void draw(sint4 frame, sint4 team) = 0;
  virtual void draw_t(sint4 frame, sint4 team) = 0;
  virtual void draw_interp(sint4 f1, sint4 f2, real4 blend, sint4 team) = 0;
  
  virtual void select(sint4 frame) = 0;
  
  virtual void draw_shadow(sint4 frame, real4 sc=0.0) = 0;
  
  sint4 num_frames() { return nframe; }

  std::map<std::string, std::pair<sint2, sint2> > name2frames;

  Model() { nframe = 0; ntag = 0; }
  virtual ~Model();

  uint4 nframe;
  uint4 ntag;
  Vec3<real4> centroid;
  real4 radius;

  Vector<BBox> bbox;
  Vector<std::string> tag_name;
  Vector<ModelTag*> tag; // each element is an array - ModelTag[nframe]
  ModelTag *all_tags;

  static void set_colors(const Vector<Vec4<real4> > &c) { colors = c; }
  const Vec4<real4>& get_color(sint4 i) { assert(i >= 0 && i < 4); return colors[i]; }

  static bool clear_fast;
  static bool verbose_load;

  struct TexPair {
    GLuint base_tex;
    GLuint team_tex;
    TexPair() { base_tex = team_tex = 0; }
  };

  TexPair *tex;

protected:
  static std::map<std::string, TexPair> textures;
  static Vector<Vec4<real4> > colors;
};

//===================================================================

/** Describes different ways of moving through an animation. */

class FrameSet
{
public:
  real4 frame;
  sint2 first;
  sint2 loop_first; // first frame after the first loop
  sint2 last;
  real4 fps;
  sint2 mode;

  sint2 f1, f2;
  real4 blend;
  
  bool is_custom;
  Vector<sint2> custom; ///< Vector of frame numbers to move between.

  enum PLAY_TYPE { STOP = 0, REPEAT = 1, MIRROR = 2, HALT = 3 };

  FrameSet() {
    is_custom = false;
  }

  bool update(real4 dt);
  void set_custom(const Vector<sint4> &v);
};

//===================================================================

class GfxAnimesh : public GfxObject
{
public:

  GfxAnimesh(GfxModule &g, const std::string &modelfile, const std::string &texfile,
             sint4 team, real4 scale, real4 dy)
  {
    set_gfxm(g);
    load(modelfile, texfile, team, scale, dy);
    shadow_tex = default_shadow_tex;
    shadow_on = 0;
    shadow_sc = 0.0;
  };

  ~GfxAnimesh();

  static void beginDraw();
  static void endDraw();

  void load(const std::string &modelfile, const std::string &texfile, sint4 team_,
            real4 scale_, real4 dy);
  void update(real4 dt);
  void draw();
  void draw_emitters();
  void draw_shape();
  void draw_bbox();
  void pick();
  Vec3<real4> midpoint() { return pose.center; }
  virtual void draw_shadow();
  void play(const std::string &, real4 fps, sint2 mode, sint2 looping = -1);
  void play(sint2 start, sint2 end, real4 fps, sint2 mode, sint2 looping = -1);
  void play_custom(const Vector<sint4> &f, real4 fps, sint2 mode = FrameSet::STOP);
  void set_idle(const std::string &, real4 fps, sint2 mode, sint2 looping = -1);
  void set_frame(const std::string &, sint2 frame);
  void set_shadow_tex(GLuint t, real4 sc = 0.0){ shadow_tex = t; shadow_sc = sc;}
  void enable_shadow(GLuint t){ shadow_on = t; }
  DrawOrder order() const { return zcat; }

  //  GLuint getTextureId() const { return mdl->tex[team]; }

  static void set_gamma(real4 g) { gamma = g; }
  static void set_mode(sint4 mm) { mode = mm; }
  static void set_texmode(sint4 mm) { texmode = mm; }
  static sint4 get_texmode() { return texmode; }
  static void set_default_shadow_tex(GLuint t){ default_shadow_tex = t; }
  static void set_texlevel(sint4 tl) { texlevel = tl; }
  static sint4 get_texlevel() { return texlevel; }

  void attachGfxObject_tag(GfxObject*, const std::string &tag);
  void attachEmitter_tag(GfxObject*, const std::string &tag);

  Model *mdl;

  struct GfxTag {
    GfxObject *obj;
    sint4 tag_num;
  };

  static ShaderProgram *default_shader;

protected:

  Vector<GfxTag> sub_objects_tag;
  Vector<GfxTag> emitters_tag;

  sint4 team;
  FrameSet idle;
  FrameSet current;
  GLuint shadow_tex;  ///< Texture to use instead of the object when drawing the shadow.
  GLuint shadow_on;   ///< Whether this object casts a shadow.
  real4 shadow_sc;    ///< Scale factor for shadow texture.

  ShaderProgram *shader;

  static std::map<std::string, Model*> name2mdl; // fixme: get rid of this static variable, delete properly
  static std::map<std::string, ShaderProgram*> name2shader;
  static real4 gamma; ///< gamma correct all model textures by this value
  static sint4 mode;  ///< The drawing mode: display list, vertex buffer, etc.
  static sint4 texmode;  ///< The texture mode.
  static sint4 texlevel; ///< Reduces texture size by a factor of 2^texlevel.
  static GLuint default_shadow_tex;
};

REGISTER_TYPEOF(260, Vector<GfxAnimesh::GfxTag>::iterator);
REGISTER_TYPEOF(261, Vector<GfxAnimesh::GfxTag>::const_iterator);

//===================================================================

// endianess conversion

static inline bool is_little() {
  sint4 e = 1;
  return *reinterpret_cast<char*>(&e);
}

inline sint4 swap_endian(sint4 x)
{
  return (x << 24) | ((x << 8) & 0xff0000) |
    ((x >> 8) & 0xff00) | ((x >> 24) & 0xff);
}

inline sint2 swap_endian(sint2 x)
{
  return (x << 8) | ((x >> 8) & 0xff);
}

inline void little2machine(sint4 &x)
{
  if (!is_little()) x = swap_endian(x);
}

inline void little2machine(real4 &x)
{
  if (!is_little()) {

    union {
      sint4 i;
      real4 f;
    };

    f = x;
    swap_endian(i);
    x = f;
  }
}

inline void little2machine(sint2 &x)
{
  if (!is_little()) x = swap_endian(x);
}

inline void little2machine(uint1 &) {}

//===================================================================

#endif
