#ifndef GfxObject_H
#define GfxObject_H

/** @file GfxObject.H
    The 3D objects that are drawn to indicate game objects and terrain.

    $Id: GfxObject.H,v 1.38 2007/04/13 15:03:44 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/gfxclient/src/GfxObject.H,v $
*/

// This is an ORTS file (c) Michael Buro, licensed under the GNU Public License

#include "GfxGlobal.H"

class GameObj;
class Game;
class GfxModule;
class Sector;
class GfxObject;
class GfxGameObject;
class GfxTerrain;
class GfxEmitter;

REGISTER_TYPEOF(200, Vector<GfxObject*>::iterator);
REGISTER_TYPEOF(201, Vector<GfxObject*>::const_iterator);
REGISTER_TYPEOF(202, Vector<GfxObject*>::reverse_iterator);
REGISTER_TYPEOF(204, std::set<GfxObject*>::iterator);
//REGISTER_TYPEOF(205, std::set<GfxObject*>::const_iterator);
REGISTER_TYPEOF(206, Vector<GfxGameObject*>::iterator);
REGISTER_TYPEOF(207, Vector<GfxGameObject*>::const_iterator);
REGISTER_TYPEOF(208, std::set<GfxGameObject*>::iterator);
//REGISTER_TYPEOF(209, std::set<GfxGameObject*>::const_iterator);

//===================================================================

class GfxObject
{
public:
  // public to allow REGISTER_TYPEOF definitions
  typedef std::map<uint4, GfxObject*> Id2GfxObjMap; 

private:
  static uint4 id0;
  static Id2GfxObjMap id2obj_map; // maps id to object (for picking)

protected:
  uint4 my_id;

public:

  static const sint4 SUB_ID_LOG = 3;      // at most (1 << SUB_ID_LOG) many sub ids per obj
  enum DrawOrder { O_TERRAIN, O_WATER, O_ON_LAND, O_IN_AIR, O_PARTICLES };
  typedef real4 R;

  struct Pose {
    Vec3<R> center;   ///< position
    real4 scale;      ///< scaling factor
    real4 y_rot;      ///< rotation along y axis
    real4 x_rot;      ///< rotation along x axis

    Pose() {
      scale = 1.0;
      y_rot = 0.0;
      x_rot = 0.0;
    }
  };

  Vec3<R> bs_c; ///< bounding sphere center
  R bs_radius;  ///< bounding sphere radius

  Vec3<R> aabb_o, aabb_d;  ///< axis aligned bounding box: origin + dimensions
  uint4 drawn;             ///< "magic", used to check whether already drawn (octree)
  Vector<Sector*> sectors; ///< octree cubes that intersect with the object

  static GfxModule *gfxm;  ///< access to options, rng etc.
  static GfxTerrain *terrain;

  Vector<GfxObject*> sub_objects;
  Vector<GfxEmitter*> emitters;
  
  DrawOrder zcat;
  bool selected;      ///< object selected
  bool dead;
  bool moved;         ///< true iff Octree needs to update this object.
  sint1 hidden;

  Pose pose;          ///< The accumulated pose.
  Vector<Pose> poses; ///< a sequence of offsets
  //  Mat4<real4> trans_mat; ///< transformation matrix

  Vec3<R> centroid;  ///< Bounding sphere center
  R radius; ///< Bounding sphere radius (from the centroid)
  

  GfxObject();
  virtual ~GfxObject();

  virtual void kill() { dead = true; };

  void attachGfxObject(GfxObject *);
  void attachEmitter(GfxEmitter *);
  
  void set_gfxm(GfxModule &m) { gfxm = &m; }
  
  void match_id(uint4 _id);
  void set_position(R x, R y, R z, uint4 p=0);
  void set_rotation(real4 y, uint4 p=0);
  void set_offset(real4 y, uint4 p=0);
  void set_pitch(real4 x, uint4 p=0);
  void set_scale(real4 s, uint4 p=0);

  void apply_pose();

  virtual DrawOrder order() const { return zcat; };
  
  virtual void update(real4 dt);
  virtual void draw_shadow() {}
  virtual void draw() = 0;
  virtual void draw_at(const Mat4<real4> &mat);
  virtual void draw_emitters();
  virtual void draw_emitters_at(const Mat4<real4> &mat);
  virtual void pick() { draw(); }
  virtual Vec3<R> midpoint() { return aabb_o+aabb_d*0.5; }

  static GfxObject* id2obj(uint4 id);
  uint4 id() const { return my_id; }
};

//===================================================================

/** Wrapper for a GfxObject that is an avatar of a game object. */

class GfxGameObject
{
public:
  // public to allow REGISTER_TYPEOF definitions
  typedef std::map<GfxObject*, GfxGameObject*> Gfx2GfxGameObj;
  typedef std::map<GfxObject*, GameObj*> Gfx2GameObj;

private:
  static Gfx2GfxGameObj gfx2gfxgame_map; // for drawing
  static Gfx2GameObj gfx2game_map; // for picking

public:
  typedef real4 R;

  real4 interp;         ///< current interpolation parameter (0-1)
  Vec3<R> center;       ///< interpolated position
  boost::array<Vec3<R>,2> position;  ///< positions of the current and previous frames
  real4 y_rot;          ///< interpolated heading
  real4 x_rot;          ///< interpolated pitch
  boost::array<real4,2> heading;     ///< heading of current and previous frame
  boost::array<real4,2> pitch;       ///< pitch of current and previous frame
  
  GameObj* const gob;
  GfxObject *gfx_obj;
  sint4 pose_num;


  GfxGameObject(GameObj *g);
  ~GfxGameObject();

  void attach_gfxobj(GfxObject *gfx);
  void game_tick();
  void update(real4 dt);
  void draw_damage();
  void draw_shape();
  
  /**
   * Determine the terrain height below the object
   * \param	noise applied to the terrain at that point
   * \return	The height of the terrain
   * \note    You should add noise to the return value if you want to
   *          find the true height as it appears visually.
   */
  real4 height_above(real4 *noise=0, real4 offset=0.f);
  static GameObj* get_gob(GfxObject *gfx);
  static GfxGameObject* get_gfx(GfxObject *gfx);
};

REGISTER_TYPEOF(241, GfxObject::Id2GfxObjMap::iterator);
REGISTER_TYPEOF(242, GfxObject::Id2GfxObjMap::const_iterator);
REGISTER_TYPEOF(243, Vector<GfxObject::Pose>::iterator);
REGISTER_TYPEOF(244, Vector<GfxObject::Pose>::const_iterator);
REGISTER_TYPEOF(245, GfxGameObject::Gfx2GfxGameObj::iterator);
REGISTER_TYPEOF(246, GfxGameObject::Gfx2GfxGameObj::const_iterator);
REGISTER_TYPEOF(247, GfxGameObject::Gfx2GameObj::iterator);
REGISTER_TYPEOF(248, GfxGameObject::Gfx2GameObj::const_iterator);
REGISTER_TYPEOF(249, Vector<GfxEmitter*>::iterator);
REGISTER_TYPEOF(250, Vector<GfxEmitter*>::const_iterator);
#endif

