#ifndef TERRAINCLASSIFIER_H
#define TERRAINCLASSIFIER_H

#include "Global.H"
#include "GameTile.H"
#include "Map.H"
#include "Game.H"
#include "Tile.H"
#include "TerrainBase.H"
#include "ScriptObj.H"
#include "GameObj.H"
#include "Object.H"
#include "Random.H"
#include <deque>
#include <stdio.h>
#include <algorithm>

namespace mapTerrain {

  typedef Vector<ScriptObj::Ptr> ObjCont;

  //---------------------------------------------------------------------
  /// \name A simple tile represents one GameTile with only terrain characteristics
  //---------------------------------------------------------------------
  struct Tile {
    enum Tile_Status {FREE, FOG, OBSTACLE, UNIT} status; ///<Indicates the status of the tile.
    sint4 freeValue;                                     ///<Indicates the distance to the nearest obstacle.
    sint4 x,y;                                           ///<The x,y coordinates of this tile, in tile units!
    
    /** A simple constructor that will zero this tile.
     */
    Tile(){ freeValue = 0; status = FOG; x = 0; y = 0; }; 
    ~Tile(){};

  };
  
  //---------------------------------------------------------------------
  /// \name The TerrainMap consists of a 2D array of tiles.
  //---------------------------------------------------------------------
  class TerrainMap {

  //---------------------------------------------------------------------
  /// \name Public Member functions.
  //---------------------------------------------------------------------

  public :

    /**   This is the default constructor.
    */
    TerrainMap(){};

    /**   This constructor will take the given game and store the current map, 
	  it will then initialize the 2D array with the same size as the game map.
    */
    TerrainMap(Game *game); 

    /**   This destructor will free all of the memory used by the grid array.
     */
    ~TerrainMap();

    /**   The print_map function will print two 2D bit arrays, the first is a 
	  grid representing the type of terrain (FREE, FOG, OBSTACLE), and the 
	  second represents a measure of how close an obstacle is to this tile.
    */
    void print_map();

    /**   This function quickly find the tile which is furthest from an obstacle in the given square.
	  @param point      The point in which we wish to find an area far away from obstacles.
	  @param location   The location of the tile farthest from obstacles.
	  @param bound      The farthest in which this search can go from the starting point
	  @param min        If we wish to find a place to build this is the minimum size we need.
	  @return           True iff a tile was found that fulfills the minimum size in the given range. 
    */
    bool find_free_loc(TerrainBase::Loc point, TerrainBase::Loc &location, sint4 bound = 64, sint4 min = 0);

    /**   This function will determine if the given area specified by a point and a range is clear of units.
	  @param point      The point in which we wish to check for units
	  @param bound      The farthest in which this search can go beyond the given point
	  @param max        The largest number of units tolerable in this area 
	  @return           True iff the area has less than or equal to max units 
    */
    bool is_free_of_units(TerrainBase::Loc point, sint4 bound, sint4 max_units = 0);

    /**   This function will find a random unexplored area near the given point
	  @param point      The point in which we wish to find an unexplored area
	  @param location   The point where there is some unexplored area
	  @param bound      The maximum distance from the given point in which we will search for an unexplored area
	  @param time_limit The maximum time alloted for this search
	  @return           True iff a valid area has been found
    */
    bool get_random_unexplored_point(TerrainBase::Loc point, TerrainBase::Loc &location, sint4 time_limit = 0, sint4 bound = 0);
 
    /**   This function will find a open location around the given game object, it assumes the game object is a building.
	  @param building   The building in which we wish to find open locations around.
	  @return           Locations around the building that are FREE, if none are found (-1,-1) is the first element of the vector returned.
    */
    Vector<TerrainBase::Loc> find_free_edge(GameObj* building);

  //---------------------------------------------------------------------
  /// \name Private functions and members
  //---------------------------------------------------------------------

  private:

    /**   This function will initialize the internal 2D map, with the given values from the game map.  
     */
    void init_blocked_tiles();

    /**   This function will compute the distances to all the obstacles in the open view.
     */
    void compute_free_areas();

    /**   This function will check if the given node is an obstacle or if it has 
	  already been visited, if not it will set the distance from the nearest 
	  obstacle and add the node to the open list of nodes to expand.
	  @param i          The x coordinate of the node.
	  @param j          The y coordinate of the node.
	  @param distance   The distance to the nearest obstacle.
    */
    void checkNode(sint4 i, sint4 j, sint4 distance);

    /**   This function will try to goto a neighboring tile that is further
	  away from obstacles.
	  @param i          The x coordinate of the tile, will be updated to the new tile.
	  @param j          The y coordinate of the tile, will be updated to the new tile.
	  @return           True iff a neighboring tile could be found that is further away from an obstacle.
    */
    bool move_one_tile(sint4 &i, sint4 &j);


    Vector<Tile> blocked_tiles;    ///<A vector containing all the tiles in which our first pass marked as obstacles 
    Tile **map;                    ///<The actual map of tiles.
    const Map<GameTile> *game_map; ///<A pointer to the game map
    sint4 height, width;           ///<The dimensions of the game map.
    sint4 tile_points;             ///<The conversion factor between tiles and coordinates
    std::deque<Tile*> openList;    ///<An open list of the tiles which still need to be expanded.
    std::deque<Tile*> tempList;    ///<Temporary list of expanded nodes
    sint4 num_players;             ///<The current number of players
    Vector<ObjCont> all_objs;      ///<All of the objects
    ObjCont buildings;             ///<Only the buildings
    ObjCont mineral_patches;       ///<The mineral patches
    ObjCont units;                 ///<All of the mobile units
    Random rand;                   ///<Random number generator (seed = 0)
    Game *game;                    ///<A pointer to the current game
  };//End of TerrainMap Clas
  
}//End of namespace mapTerrain

REGISTER_TYPEOF(3400, Vector<mapTerrain::Tile>::iterator);
REGISTER_TYPEOF(3401, Vector<mapTerrain::Tile>::const_iterator);
REGISTER_TYPEOF(3402, std::deque<mapTerrain::Tile*>::iterator);
REGISTER_TYPEOF(3403, std::deque<mapTerrain::Tile*>::const_iterator);

#endif /*TERRAINCLASSIFIER.H*/
