#include "OgreGfxModule.H"

OgreGfxModule::OgreGfxModule(){

  //First Ogre command ever run in an Ogre app; create the application's root object
  mRoot = new Ogre::Root("apps/ortsgo/scripts/plugins.cfg","apps/ortsgo/scripts/ogre.cfg" );
  //(The two strings tell Ogre where to find the plugin and config scripts)

  init_setupResources();//create references to all our meshes textures, and other stuff we'll need

  if (!init_configure()){  //Show the Ogre configuration dialog box, but terminate app if it fails to open
    std::cout<<"Error! Ogre Configuration failure!\n Exiting now..."<<std::endl;
    exit(1);
  }

  // Create the SceneManager (efficently manages all objects in the program), in this case a generic one
  //mSceneMgr = mRoot->createSceneManager(Ogre::ST_GENERIC, "ExampleSMInstance");
  mSceneMgr = mRoot->createSceneManager( Ogre::ST_EXTERIOR_CLOSE ); //updated for 1.2  CHANGED TO SUPPORT WORLD GEOMETRY
  
  init_createCamera();//create a camera and choose its position
  init_createViewports();//create any viewports to view from

  // Set default level of detail to use on textures (i.e. "mipmaps")
  Ogre::TextureManager::getSingleton().setDefaultNumMipmaps(5);
  //parse those resource scripts now so that you don't have to do it at runtime (here just for efficeincy)
  Ogre::ResourceGroupManager::getSingleton().initialiseAllResourceGroups();
  
  //Create a frame listener that will detect mouse clicks and keyboard buttons gettin' pressed
  init_createFrameListener();

  //Finally! Let's start placing objects and stuff into the scene!
  createScene();

  return;
}

//Gets a reference to all resources that this program is going to need (listed in "resources.cfg")
void OgreGfxModule::init_setupResources(){
	
  // Load resource paths from config file
  Ogre::ConfigFile cf;
  cf.load("apps/ortsgo/scripts/resources.cfg");//TODO: let this be changed at runtime

  // Go through all sections & settings in the file
  Ogre::ConfigFile::SectionIterator seci = cf.getSectionIterator();

  Ogre::String secName, typeName, archName;
  while (seci.hasMoreElements()){
    secName = seci.peekNextKey();
    Ogre::ConfigFile::SettingsMultiMap *settings = seci.getNext();
    Ogre::ConfigFile::SettingsMultiMap::iterator i;
    for (i = settings->begin(); i != settings->end(); ++i){
      typeName = i->first;
      archName = i->second;
      Ogre::ResourceGroupManager::getSingleton().addResourceLocation(archName, typeName, secName);
    }
  }
}

//Returns true so long as the user did not cancel the loading of the configuration dialog box
bool OgreGfxModule::init_configure(){
  // Show the configuration dialog and initialise the system
  // You can skip this and use root.restoreConfig() to load configuration
  // settings if you were sure there are valid ones saved in ogre.cfg
  if(mRoot->restoreConfig() || mRoot->showConfigDialog()){  //OK button = true, Cancel button = false
    mWindow = mRoot->initialise(true);//user pressed "OK"
    return true;
  }
  return false;//User cancelled
}

//get our camera all set up
void OgreGfxModule::init_createCamera(){
  mCamera = mSceneMgr->createCamera("viewingCamera");// Create the camera
  mCamera->setPosition(Ogre::Vector3(512,900,1024));// Position it at 100 in Z direction
  mCamera->lookAt(Ogre::Vector3(512,0,512));//have it look at the origin
  mCamera->setNearClipDistance(5);//clip anything less than 5 units close to the camera
}

//Create our viewports (in this case only 1 viewport that uses up the whole screen)
void OgreGfxModule::init_createViewports(){
  Ogre::Viewport* vp = mWindow->addViewport(mCamera); //Create one viewport that covers the entire window
  vp->setBackgroundColour(Ogre::ColourValue(0,0,0)); //set the background color of the viewport to black
  // Alter the camera aspect ratio to match the viewport
  mCamera->setAspectRatio(Ogre::Real(vp->getActualWidth()) / Ogre::Real(vp->getActualHeight()));
}

void OgreGfxModule::createScene(){
  //Firstly, set up some ambient lighting, and choose the really expensive (but accurate) shadow technique.
  mSceneMgr->setAmbientLight( Ogre::ColourValue( .25, .25, .25 ) );
  //mSceneMgr->setShadowTechnique( Ogre::SHADOWTYPE_TEXTURE_MODULATIVE );

  //Next, add a light source so that we can see all the pretty shadows :)
  Ogre::Light* light = mSceneMgr->createLight( "Light1" );
  light->setType( Ogre::Light::LT_DIRECTIONAL);
  light->setDirection( Ogre::Vector3( -0.3f,-0.7f,0.0f ) );

  //Ogre::Entity* lightSource = mSceneMgr->createEntity("light source", "robot.mesh");
  //Ogre::SceneNode* lightNode = mSceneMgr->getRootSceneNode()->createChildSceneNode("light node");
  //lightNode->attachObject(lightSource);
  //lightNode->attachObject(light);
  //lightNode->setPosition(Ogre::Vector3( 0.0f, (8.0f*30.0f), (8.0f*60.0f) ) );
  
  Ogre::Plane skyPlane;
  skyPlane.d = 1000;
  skyPlane.normal = Ogre::Vector3::NEGATIVE_UNIT_Y;
  mSceneMgr->setSkyPlane( true, skyPlane, "Examples/SpaceSkyPlane", 1500, 5);

  ObjectManager::getSingleton().mech = new OgreMD3Model();
  ObjectManager::getSingleton().mech->load("apps/ortsgo/ortsOgreMedia/mech.md3");
  ObjectManager::getSingleton().mech->createMD3Object(mSceneMgr, "mech");

  //mSceneMgr->setWorldGeometry( "terrain.cfg" );
}

void OgreGfxModule::init_createFrameListener(){
  mInputDevice = Ogre::PlatformManager::getSingleton().createInputReader();
  mFrameListener= new OgreGfxModuleFrameListener(mWindow, mCamera, mSceneMgr);
  mFrameListener->showDebugOverlay(true);//Show all that good stuff like fps stats
  mRoot->addFrameListener(mFrameListener);//add the new frameListener to recieve messages
  mInputDevice->initialise(mWindow);//making sure that input reading is initiallzed
}

//obtain a reference to the game's client state and the game state module
bool OgreGfxModule::init(GameStateModule* gameStateModule){
  gsm = gameStateModule;
  mFrameListener->init(gameStateModule);
  return true;
}

//Renders a single frame. Returns false if the user intends to quit,
//and returns true otherwise
bool OgreGfxModule::renderOneFrame(){
  
  Ogre::PlatformManager::getSingleton().messagePump(mWindow);

  if(mRoot->renderOneFrame() == false)//try to render a frame
    return false;//exit if the rendering failed

  //if (mInputDevice->isKeyDown(Ogre::KC_ESCAPE))//if the user pressed ESC...
  //  return false;//...then indicate that it's time to terminate the program
 
  return true;
}

//Tells the FrameListener to process changes
void OgreGfxModule::process_changes(){
  mFrameListener->process_changes();
  return;
}

//Free up all allocated memory
OgreGfxModule::~OgreGfxModule(){
  if (mFrameListener)
    delete mFrameListener;
  mFrameListener=NULL;

  if (mRoot != NULL)
    delete mRoot;
  mRoot = NULL;
}
