
#include "Global.H"

#include "SI_SimulationState.h"

#include "GameStateModule.H"
#include "Game.H"
#include "GameObj.H"

#include "SI_Order.h"

namespace Simul{

  SimulationState::SimulationState(const SimulationState& b){
    FORALL(b.players,it){
      players[it->first] = (it->second)->clone();
    }
    groupSize = 50;
    smartJoinSize = 5;
    smartJoin = false;
    useMoveForAttack = false;
    findspot = false;
  }

  SimulationState::~SimulationState(){
    cleanup();
  }

  SimulationState& SimulationState::operator= (const SimulationState& b){
    if (this == &b) return *this;

    cleanup();

    FORALL(b.players,it){
      players[it->first] = (it->second)->clone();
    }

    return *this;
  }

  SimulationState* SimulationState::clone(){
    return new SimulationState(*this);
  }

  void SimulationState::cleanup(){
    FORALL(players,it){
      delete it->second;
      it->second = NULL;
    }
    players.clear();
  }

  GameObj* SimulationState::getNearestEnemyUnitOfGroup(uint4 playerID,const ObjectVector& objects){
    real8 nearest = 2e8;
    GameObj* nearestUnit = NULL;

    FORALL(players,it) {
      if((it->second)->getID() == playerID) continue; // don't check our own units
      FORALL((it->second)->getOrtsUnits(),it2){
	FORALL(objects,it3){
	  real8 dist = (*it3)->distance_to(*(it2->first));
	  if(dist < nearest) {
	    nearestUnit = it2->first;
	    nearest = dist;
	  }
	}
      }
    }
    return nearestUnit;
  }

  GameObj* SimulationState::getAverageNearestEnemyUnitOfGroup(uint4 playerID,const ObjectVector& objects){
    real8 nearest = 2e8;
    GameObj* nearestUnit = NULL;

    FORALL(players,it) {
      if((it->second)->getID() == playerID) continue; // don't check our own units
      FORALL((it->second)->getOrtsUnits(),it2){
        real8 totalDistance = 0.0;
	FORALL(objects,it3){
	  totalDistance += (*it3)->distance_to(*(it2->first));
	}
        if(totalDistance < nearest){
          nearest = totalDistance;
          nearestUnit = it2->first;
        }
      }
    }
    return nearestUnit;
  }

  GameObj* SimulationState::getNearestEnemyBaseOfGroup(uint4 playerID,const ObjectVector& objects){
    real8 nearest = 2e8;
    GameObj* nearestBase = NULL;

    FORALL(players,it) {
      if((it->second)->getID() == playerID) continue; // don't check our own units
      FORALL((it->second)->getOrtsBases(),it2){
	FORALL(objects,it3){
	  real8 dist = (*it3)->distance_to(*(it2->first));
	  if(dist < nearest) {
	    nearestBase = it2->first;
	    nearest = dist;
	  }
	}
      }
    }
    return nearestBase;
  }


  TerrainBase::Loc SimulationState::getCenterOfObjects(const ObjectVector& objects){

    sint4 maxX = min_sint4;
    sint4 maxY = min_sint4;
    sint4 minX = max_sint4;
    sint4 minY = max_sint4;

    FORALL(objects, it) {
      sint4 x,y;

      (*it)->get_center(x,y);

      if (x > maxX) maxX = x;
      if (y > maxY) maxY = y;
      if (x < minX) minX = x;
      if (y < minY) minY = y;
    }

    return TerrainBase::Loc((maxX + minX)/2, (maxY+minY)/2);
  }

  GameObj* SimulationState::getNearestObject(const ObjectVector& objects,const TerrainBase::Loc& loc){
    if(objects.empty()) return NULL;
    real8 closestDistance = 2e8;
    GameObj* closestObject = NULL;
    FORALL(objects,it){
      sint4 x,y;
      (*it)->get_center(x,y);
      real8 distance = sqrt(pow((x - loc.x),2) + pow((y - loc.y),2));
      if(distance < closestDistance){
        closestDistance = distance;
        closestObject = *it;
      }
    }
    return closestObject;
  }

  GameObj* SimulationState::getNearestObject(const ObjectVector& objects,GameObj* obj){
    if(obj == NULL) ERR("can't find nearest group to a NULL group");
    sint4 x,y;
    obj->get_center(x,y);
    return getNearestObject(objects,TerrainBase::Loc(x,y));
  }

  void SimulationState::getNearestObjectsToObject(ObjectVector& closestObjects,
    ObjectVector originalObjects,GameObj* obj,uint4 numObjects){

      // for more robustness, if we want more objects than there are available in the vector, just select all we can.
    if(numObjects > originalObjects.size()){
      //ERR("can't select more objects than we have in our vector!");
      numObjects = originalObjects.size();
    }
    if(numObjects <= 0) return;

    closestObjects.clear();

    while(closestObjects.size() < numObjects){
      GameObj* mostNearestGroup = getNearestObject(originalObjects,obj);
      closestObjects.push_back(mostNearestGroup);
      removeObjectFromVector(mostNearestGroup,originalObjects);
    }

  }


  bool SimulationState::removeObjectFromVector(GameObj* obj, ObjectVector& vector){
    FORALL(vector,it){
      if(*it == obj){
        vector.erase(it);
        return true;
      }
    }
    return false;
  }

  bool SimulationState::subtractVector(const ObjectVector& vectorToSubtract,ObjectVector& vector){
    bool succesfullyRemovedAllObjects = true;
    FORALL(vectorToSubtract,it){
      if(!removeObjectFromVector(*it,vector)){
        succesfullyRemovedAllObjects = false;
      }
    }
    return succesfullyRemovedAllObjects;
  }

  bool SimulationState::enemyInRange(Group* group){
    Player* owningPlayer = getPlayer(group->getPlayerID());
    assert(owningPlayer != NULL);

    FORALL(players,playerIterator){
      if(owningPlayer->getID() == (playerIterator->second)->getID()) continue; // don't check your own objects...

      BaseVector* playerBases =  (playerIterator->second)->getBases();

      FORALL(*playerBases,it){
        if((*it)->withinRange(group)){
	  return true;
	}
      }
      GroupVector* playerGroups =  (playerIterator->second)->getGroups();

      FORALL(*playerGroups,it){
        if((*it)->withinRange(group)){
	  return true;
	}
      }
    }
    return false;
  }

  Player* SimulationState::getPlayer(uint4 playerID){
    PlayerMap::iterator it;
    it = players.find(playerID);
    if(it != players.end()){
      return it->second;
    }else{
      return NULL;
    }
  }

  void SimulationState::updateObjects(real8 lastMoveUpdateMillis, real8 currUpdateMillis){

    moveObjects(currUpdateMillis - lastMoveUpdateMillis);
    resolveCollisions(currUpdateMillis);

    // now we delete any objects that have < 0 hitpoints
    //deleteDeadObjects();

    updateOrders();

    updateStrategies();
  }

  TaskCreateFunctionPtr SimulationState::findBestStrategy(
    uint4 playerID,
    real8 currentTime,
    real8 allowedComputationTime,
    real8 timeToSimulate,
    SimulationResult& simResult)
  {
    simResult.isDone = false;

    //double simStartTime = (double)glutGet(GLUT_ELAPSED_TIME);
    real8 simStartTime = getCurrentTime();

    TaskCreateFunctionPtr bestStrategy = simResult.bestStrategy;
    real8 bestTotalValue = simResult.bestValue;
    real8 currentTotalValue = simResult.currentValue;

    for(uint4 ours = simResult.ourStrategyIndex; ours < NUM_STRATEGIES; ++ours){

      for(uint4 theirs = simResult.theirStrategyIndex; theirs < NUM_STRATEGIES; ++theirs){

	// before doing any simulation.. check if we still have some computation time left to do it...
	// if(((double)glutGet(GLUT_ELAPSED_TIME) -simStartTime) > allowedComputationTime){
	if((getCurrentTime() - simStartTime) > allowedComputationTime){

	  // if we don't have any time left save the current state
	  simResult.ourStrategyIndex = ours;
	  simResult.theirStrategyIndex = theirs;
	  simResult.bestValue = bestTotalValue;
	  simResult.currentValue = currentTotalValue;
	  simResult.bestStrategy = bestStrategy;

	  return NULL;
	}

	SimulationState* tempState = this->clone();

	// now we copy the tasks for the players IFF their current task matches the task we will be simulating
	FORALL(players,it){
	  // NOTE: since tasks and orders are not copied by default, we copy them here when we need to
	  if((it->second)->getID() == playerID){
	    if((it->second)->getStrategy() != NULL && (it->second)->getStrategy()->getFunction() == strategyList[ours]){
	      copyPlayerInfo(it->second,tempState);
	    }
	  }else{
	    if((it->second)->getStrategy() != NULL && (it->second)->getStrategy()->getFunction() == strategyList[theirs]){
	      copyPlayerInfo(it->second,tempState);
	    }
	  }
	}

	// for now, count up the total value of this strategy vs ALL of theirs
	real8 combinationValue =  tempState->getStrategyValue(playerID, strategyList[ours],strategyList[theirs],currentTime,timeToSimulate);

	currentTotalValue +=combinationValue;

	delete tempState;

        /*
      PRINT(boost::format("ours:[%u] %p theirs:[%u] %p value: %lf ") % ours % (void*)strategyList[ours] % theirs % (void*)strategyList[theirs] % combinationValue);
        */
      }

        //PRINT(boost::format("-->total value of strategy [%u] %p  is: %lf") % ours % (void*)strategyList[ours] % currentTotalValue);


      if(currentTotalValue > bestTotalValue){
	bestTotalValue = currentTotalValue;
	bestStrategy = strategyList[ours];
      }

      currentTotalValue = 0.0;
      // after the first saved iteration, we reset this.  only first iteration needs to be resumed afterall
      simResult.theirStrategyIndex = 0;
    }

    // if we are done, reset all the simulation values to their
    // defaults... and then return the best strategy
    simResult.isDone = true;
    simResult.ourStrategyIndex = 0;
    simResult.theirStrategyIndex = 0;
    simResult.bestValue = -2e30;
    simResult.currentValue = 0.0;
    simResult.bestStrategy = NULL;

      //PRINT(boost::format("---->best in search: %p\n") % (void*)bestStrategy);
    return bestStrategy;
  }

  real8 SimulationState::advanceTime(real8 lastTime){
    real8 lastMoveUpdateMillis = lastTime;
    lastTime = getNextTime(lastTime);

    // this can happen sometime, so when next time is = to current time,
    // make sure we advance at least a little bit

    //FIXME we shouldn't need this, get rid of it if possible
    if(lastTime == lastMoveUpdateMillis){
      lastTime+=50;
    }

    updateObjects(lastMoveUpdateMillis,lastTime);

    return lastTime;
  }

  bool SimulationState::isGameOver(){
    FORALL(players,it){
      if((it->second)->getVictoryValue() <= 0.0 || (it->second)->getGroups()->size() == 0)
	return true;
    }
    return false;
  }

  bool SimulationState::hasWon(uint4 playerID){
    FORALL(players,it){
      if((it->second)->getID() == playerID) continue;

      if((it->second)->getVictoryValue() > 0.0){
	return false;
      }
    }
    return true;
  }

  // FIXME for now, just return the total value of our bases... want to make this much more complex!
  real8 SimulationState::evaluate(uint4 playerID){
    real8 value = 0.0;
    FORALL(players,it){
      real8 playerValue = (it->second)->getVictoryValue();
      //if(playerValue <= 0.0) playerValue -= VICTORY_BONUS; // if player has no bases left, we have to make our evaluation very low!!!
      //if((it->second)->getGroups()->size() == 0) playerValue -= VICTORY_BONUS*10.0; // FIXME this is just arbitrary right now

      if((it->second)->getID() == playerID){
	value+=playerValue;
      }else{
	value-= playerValue;
      }
    }
    return value;
  }

  real8 SimulationState::getNextTime(real8 currentTime){
    real8 minTime = 2e30;

    real8 collisionTime = getNextCollisionTime(currentTime);
    if(collisionTime < minTime) minTime = collisionTime;

    real8 orderCompletionTime = getNextOrderCompletionTime(currentTime);
    if(orderCompletionTime < minTime) minTime = orderCompletionTime;

    //real8 shootingTime = getNextShootingTime(currentTime);
    //if(shootingTime < minTime) minTime = shootingTime;

    //real8 timeoutTime = getNextOrderTimeoutTime(currentTime);
    //  if(timeoutTime < minTime) minTime = timeoutTime;

    if(minTime == 2e30){
      return currentTime;
    }

    if(minTime < currentTime){
    //print("error, mintime of " + convertToString(minTime) + " is less than the currentTime of: " + convertToString(currentTime));
      return currentTime;
    }

    //print("new minTime = " + convertToString(minTime));

    // IMPORTANT: the +0.01 is just to prevent some rounding off problems which can be inefficient, since some units don't quite reach their goal sometimes
    // it's equivalent to 1/100 milliseconds so not significant at all
    return minTime;
  }

  void SimulationState::updateOrders(){
    FORALL(players,it){
      (it->second)->updateOrders(*this);
    }
  }

  void SimulationState::moveObjects(real8 timeElapsed){
    FORALL(players,it){
      (it->second)->executeOrders(timeElapsed);
    }
  }

void SimulationState::resolveCollisions(real8 /*currentTime*/){
    // yes,, it is n^2, and could be made a bit better by doing both
    // evaluations for collisions, but total computing time is around
    // the same...(i think)

    FORALL (players, currentPlayerIterator) {

      FORALL (*(currentPlayerIterator->second)->getGroups(), it) {
	Group* currentGroup =(*it);

        Vector<Base*> basesInRange;
        Vector<Group*> groupsInRange;

	FORALL(players,secondPlayerIterator){
	  if((currentPlayerIterator->second)->getID() == (secondPlayerIterator->second)->getID()){
	    continue;
	  }

	  FORALL(*(secondPlayerIterator->second)->getBases(), it2) {
	   if((*it2)->withinRange(currentGroup)){
  	     basesInRange.push_back(*it2);
  	     //basesInRangeMap[*k].push_back(it2->second);
	   }
	  }

	  FORALL(*(secondPlayerIterator->second)->getGroups(), it2) {
	   if((*it2)->withinRange(currentGroup)){
  	     groupsInRange.push_back(*it2);
  	     //groupsInRangeMap[*k].push_back(it2->second);
	   }
	  }
	}
        // if no objects in range, we can continue
	if(basesInRange.empty() && groupsInRange.empty()) continue;

        /*
        // if we are targetting bases first...
	if(currentGroup->getBaseFirstAttackPriority()){
	  if(attackBasesInRange(basesInRangeMap,currentGroup,currentTime)){

	  }else if(attackUnitsInRange(groupsInRangeMap,currentGroup,currentTime)){

	  }else{
	// no targets found, should never get here!
	// print("no targets found in collisions while trying to deal damage, this should NOT happen!");
	  }
	}else{
	  if(attackUnitsInRange(groupsInRangeMap,currentGroup,currentTime)){

	  }else if(attackBasesInRange(basesInRangeMap,currentGroup,currentTime)){

	  }else{
	// no targets found, should never get here!
	//print("no targets found in collisions while trying to deal damage, this should NOT happen!");
	  }
	}
        */
      }
    }
  }

    // function originally from Collision2D.C from ORTS

  /** Solves \f$ at^2 + bt = c; a > 0 \f$
	@return true iff solution exists

  robust solution (numerical recipies)
  */

  bool SimulationState::solveQuadratic(real8 a, real8 b, real8 c, real8& t1, real8& t2){

    if (a <= 0.0){
      //print("ERROR: Trying to solve quadratic with a <=0");
      return false;
    }
    // no need to do this, pass a negative c instead
    //c = -c;
    real8 squareRootPart = b*b-4.0*a*c;

    // can't have a negative square root, so return false if this is negative
    if (squareRootPart < 0.0){
     //print("error, squareRootPart is negative: "+convertToString(squareRootPart));
      return false;
    }
    // robust root computation (Numerical Recipies p.184)

    real8 q = -0.5L * (b + sgn(b) * sqrt(squareRootPart));
    if(q == 0.0){
      //print("error in quadratic, q was 0");
      return false;
    }
    t1 = q / a;
    t2 = c / q;

    if (t1 > t2){
      real8 temp = t2;
      t2 = t1;
      t1 = temp;
    }
     //assert(!(t1 > t2));

    real8 e1 = fabs(a*t1*t1+b*t1+c);
    real8 e2 = fabs(a*t2*t2+b*t2+c);
    if (e1 > 1.5e-5) ERR2("error in quadratic: ",e1);
    if (e2 > 1.5e-5)  ERR2("error in quadratic: ",e2);

    return true;
  }

  // originally from Collision2D.H from ORTS

  /** Compute collision time of a (moving) circle and a (moving) point.
      Used to check for collision with the endpoint of a line.
      @param circlePos     circle center
      @param radius    square of the circle's radius
      @param pointPos     point's location
      @param relativeHeading     circle's heading relative to the point
      @param startTime  start simulation at this time
  */


  real8 SimulationState::getCirclePointCollisionTime(const Vector3D& circlePos,
    real8 radius,
    const Vector3D& pointPos,
    const Vector3D& relativeHeading,
    real8 startTime)
    {

    Vector3D positionDifference = circlePos - pointPos;
    real8 quadraticValueC =
      (radius*radius) -
      (positionDifference.x*positionDifference.x) -
      (positionDifference.y*positionDifference.y);

    //int c_h = c_p * h; // c dot h
    real8 differenceHeading = positionDifference.dot(relativeHeading);

    // if they are heading away from each other and not touching, then
    // no need to do any more work...

    if (differenceHeading > 0.0) {   // moving away from each other
      if (quadraticValueC >= 0.0) { // they don't intersect
        //print("moving from each other");
	return -1.0;
      }
    }

    real8 t1= 0.0;
    real8 t2 = 0.0;

    real8 quadraticValueA =
      (relativeHeading.x*relativeHeading.x) + (relativeHeading.y*relativeHeading.y);
    real8 quadraticValueB = 2.0 * differenceHeading;
    bool isCollision =
      solveQuadratic(quadraticValueA,quadraticValueB,-quadraticValueC, t1, t2); // (a)t^2 + (b)t - c = 0

    if (!isCollision){
      //print("no collision");
      return -1.0;
    }

    t1+=startTime;
    t2+=startTime;
    if (t1 == t2){
       //print("t1 and t2 are equal!");
      return -1.0; // point intersects tangentially
    }

    if (t2 <= startTime){
      //print("t2 <= starttime");
      return -1.0;
    }
    if (t1 == startTime){
      //print("WARNING: time is equal to startTime");
      return -1.0;
     //return startTime; // intersecting at time
    }
    if(t1 < startTime || t2 < startTime){
     //print("WARNING: time of collision is less than start time, this is wrong!");
     //return startTime;
      return -1.0;
    }

    return t1;
  }

  real8 SimulationState::getNextCollisionTime(real8 currentTime){
    real8 minimumTime = 2e30;

    FORALL(players,currentPlayerIterator){
      GroupVector* currPlayerGroups =  (currentPlayerIterator->second)->getGroups();

      FORALL(*currPlayerGroups,it){
        // FIXME later make this work for bases as well
	Group* currentObject = (*it);

	Vector3D currentObjectHeading;
	if(currentObject->getOrder() == NULL){
	  currentObjectHeading = Vector3D(0.0,0.0,0.0);
	}else{
	  currentObjectHeading =
	    (currentObject->getOrder()->getDestination()-currentObject->getPosition()) |
	    currentObject->getOrder()->getSpeed();
	}

	FORALL(players,otherPlayerIterator){
      // don't check your own objects...
	  if(currentPlayerIterator == otherPlayerIterator) continue;

	  GroupVector* otherPlayerGroups =
	    (otherPlayerIterator->second)->getGroups();

	  FORALL(*otherPlayerGroups,it2){
	    Group* otherObject = (*it2);

	    // if objects are already within each other's range, don't do them anymore.

	    if((currentObject->getPosition() - otherObject->getPosition()).length() <=
		currentObject->getExtents() + currentObject->getMaxAttackRange() +
		otherObject->getExtents()) {
	           //print("inside object!!!");
		  continue;
		}

	    Vector3D otherObjectHeading;

	    if(otherObject->getOrder() == NULL){
	      otherObjectHeading = Vector3D(0.0,0.0,0.0);
	    }else{
	      otherObjectHeading =
		(otherObject->getOrder()->getDestination()-otherObject->getPosition()) |
		otherObject->getOrder()->getSpeed();
	    }

	    Vector3D relativeHeading = currentObjectHeading - otherObjectHeading;

	    real8 collisionTime =
	      getCirclePointCollisionTime(currentObject->getPosition(),
					  currentObject->getExtents()+
					  currentObject->getMaxAttackRange()+
					  otherObject->getExtents(),
					  otherObject->getPosition(),
					  relativeHeading,
					  currentTime);

	     //print("found time: "+convertToString(collisionTime));
	    if(collisionTime == -1.0){
	      continue;
	    }

	    if(collisionTime < minimumTime) minimumTime = collisionTime;
	  }
	}
      }
    }
    return minimumTime;
  }

  real8 SimulationState::getNextOrderCompletionTime(real8 currentTime){
    real8 minCompletionTime = 2e30;

    FORALL(players,playerIterator){
      GroupVector* currPlayerGroups =  (playerIterator->second)->getGroups();

      FORALL(*currPlayerGroups,it){
	Group* currentGroup = (*it);

	if(currentGroup->getOrder() == NULL) continue;

        // FIXME not sure if the next line should be left in or not.
	if(currentGroup->getPosition() == currentGroup->getOrder()->getDestination())
	  continue;

        //print("unit " + currentGroup->getName() + " order start time: "+convertToString(currentTime));
	real8 estimatedTime = currentTime +
	  (currentGroup->getOrder()->getDestination() - currentGroup->getPosition()).length() /
	  ((real8)currentGroup->getOrder()->getSpeed());

	if(estimatedTime == currentTime) continue;

        //print("unit " + currentGroup->getName() + " order estimated end time: "+convertToString(estimatedTime));

	if(estimatedTime < minCompletionTime){
	  minCompletionTime = estimatedTime;
	}
      }
    }
   return minCompletionTime;
  }

  real8 SimulationState::getStrategyValue(
				      uint4 playerID,
				      TaskCreateFunctionPtr ourStrategy,
				      TaskCreateFunctionPtr theirStrategy,
				      real8 currentTime,
				      real8 timeToSimulate)
  {

    FORALL (players, it) {
      ObjectVector newVector;

      // by default, give ALL the units of a player the same strategy, this can be changed


      FORALL ((it->second)->getOrtsUnits(), it2) {
	newVector.push_back(it2->first);
      }

      if(newVector.empty()){
	(it->second)->setStrategy(NULL);
	continue; // don't do anything if the player has no more groups
      }

      if((it->second)->getID() == playerID){
	if(ourStrategy == NULL) continue;
	if(((it->second)->getStrategy() == NULL || (it->second)->getStrategy()->getFunction() != ourStrategy)){
  	(it->second)->setStrategy(ourStrategy((it->second)->getID(),this,newVector,0.0));
	}
      }else{
	if(theirStrategy == NULL) continue;
	if(((it->second)->getStrategy() == NULL || (it->second)->getStrategy()->getFunction() != theirStrategy)){
  	(it->second)->setStrategy(theirStrategy((it->second)->getID(),this,newVector,0.0));
	}
      }
    }

    real8 value = 0.0;
    value =  simulateForTime(playerID,currentTime,timeToSimulate);

    return value;
  }

  real8 SimulationState::simulateForTime(
				      uint4 playerID,
				      real8 currentTime,
				      real8 timeToSimulate)
  {
    real8 timeToSimulateTo = currentTime+timeToSimulate;
    real8 nextSimulationTime = currentTime;

    while(nextSimulationTime < timeToSimulateTo && !isGameOver()){
      nextSimulationTime = advanceTime(nextSimulationTime);
    }

    return evaluate(playerID);
  }


  void SimulationState::updateStrategies(){
    FORALL(players,it){
      Task* currStrategy = (it->second)->getStrategy();
      if(currStrategy != NULL){
	bool isDone = currStrategy->update();

	if(isDone){
  	   (it->second)->setStrategy(NULL);
	}
      }
    }
  }

  void SimulationState::copyPlayerInfo(Player* origPlayer,SimulationState* tempState){
    Player* tempPlayer = tempState->getPlayer(origPlayer->getID());
    // first copy the task
    tempPlayer->setStrategy(origPlayer->getStrategy()->clone(tempState,NULL));
    // then copy the orders
    /*
    FORALL(*origPlayer->getGroups(),it2){
      Group* tempGroup = tempPlayer->getGroup((it2->second)->getID());

      Order* currOrder = (it2->second)->getOrder();
      if(currOrder != NULL){
	tempGroup->setOrder(currOrder->clone());
      }
    }
    */
  }

}

