/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Global.H"
#include "RT_ClientEnv.H"
#include "RTS_Game.H"
#include "System.H"
#include "Server.H"
#include "ServerClient.H"
#include "RT_Client.H"
#include "RT_GameParam.H"

const sint4 USLEEP = 10000;

#define CMDERR(x) { cout << "ERR: " << x << endl; continue; }

RT_ClientEnv::RT_ClientEnv(RTS_Options &opt_) : opt(opt_), compress(8)
{
  state = ST_IDLE;
  rt_client = 0;
  last_event = EVT_NONE;
  client = 0;
  game = 0;
  client_died = false;
  first_obj = true;
  local = false;
  opt.get("-cpind", cpind);
}
  

RT_ClientEnv::~RT_ClientEnv() {
  if (rt_client) delete rt_client;
  if (client) delete client;
  if (game) delete game;
}


RT_ClientEnv::Event RT_ClientEnv::get_event()
{
  if (last_event == EVT_STOP) return last_event;
  if (client_died) { last_event = EVT_STOP; return last_event; }
  
  last_event = EVT_NONE;

  string msg, act;
  sint4 ret = client->read_msg(msg);
  if (ret <  0) { client_died = true; last_event = EVT_STOP; return last_event; }
  if (ret == 0) { last_event = EVT_NONE; return last_event; }

  // received message
  
  if (msg.size() == 0) ERR("empty message");
  
  switch (state) {
    
  case ST_RECV_OBJS:

    // receive all new objects

    FOREVER {

      // assume: object messages don't start with '#'!
      
      if (msg[0] == '#') { // end of block

	game->update_view(decompress, obj_msgs);
	game->prop_vectors(objs);
	game->clear_actions();
	if (local) game->create_vicinities(cpind, vics);

	istringstream iss(msg);
	string s;
	sint4 winner;
	iss >> s; // skip #
	iss >> winner;
	if (iss) {

	  // game end

	  if (game) {
	    bool d;
	    opt.get("-disp", d);
	    if (d) game->display();
	  }

	  if (winner < 0)
	    cout << "it's a draw" << endl;
	  else if (winner == cpind)
	    cout << "you win" << endl;
	  else
	    cout << "you lose" << endl;
	    
	  state = ST_IDLE;
	  last_event = EVT_STOP;
	  return last_event;      

	} else {

	  state = ST_RECV_OBJS;
	  first_obj = true;
	  last_event = EVT_NEW_OBJS;
	  return last_event;
	    
	}
	
      } else {

	if (first_obj) obj_msgs.clear();
	obj_msgs.push_back(msg);
	first_obj = false;
	
	// next message
	
	sint4 ret = client->read_msg(msg);
	if (ret < 0) {
	  client_died = true; last_event = EVT_STOP;
	  return last_event;
	}
	if (ret == 0) break;
      }
    }
    
  default:
      break;
  }	  

  last_event = EVT_NONE;
  return last_event;
}


bool RT_ClientEnv::set_action(uint4 id, Action &act)
{
  if (game->set_action(cpind, id, act)) ERR("can't set action!");
  return false;
}

void RT_ClientEnv::send_actions()
{
  vector<string> msgs;
  
  game->create_action_messages(compress, cpind, msgs);

  FORU (i, msgs.size()) {
    if (client->write_msg(msgs[i], true) <= 0) {
      client_died = true;
      break;
    }
  }

  String m;
  m.form("# %d", game->elems_checksum());
  if (client->write_msg(m, true) <= 0) client_died = true;
  client->write_flush();

  if (game) {
    bool d;
    opt.get("-disp", d);
    if (d) game->display();
  }
}


void RT_ClientEnv::run()
{
  Server server(131072, 131072);
  sint4 cport;
  opt.get("-cport", cport);

  if (cport <= 1023) ERR("cport <= 1023");
    
  server.init_listen(cport);

  string cpact;
  opt.get("-cpact", cpact);
  cout << "client listening at port: " << cport << endl;
  cout << "client action: " << cpact << endl;
  
  state = ST_IDLE;
  client_died = false;
  first_obj = true;
  
  FOREVER {

    if (client && client_died) {
      REM("broken connection");
      delete client;
      client = 0;

      if (server.clients.size() != 1) ERR("clients.size != 1");
      
      server.clients.pop_front();

      state = ST_IDLE;
      client_died = false;
      first_obj = true;
    }

    switch (state) {

    case ST_IDLE:

      // wait for connection

      // cout << "." << flush;
      
      if (server.handle_connection()) {
	cout << "connected" << endl;
	client = server.clients.front();
	state = ST_CREATE;
	client_died = false;
	compress.init();
	decompress.init();
      }
      usleep(USLEEP);
      break;

    case ST_CREATE:

      {
	// wait for create message
	
	string msg;
	sint4 ret = client->read_msg(msg);
	
	if (ret < 0) { client_died = true; break; }
	
	if (ret > 0) {

	  // received message

	  istringstream iss(msg);

	  string cmd;
	    
	  iss >> cmd;
	  if (!iss || cmd != "create") {
	    cerr << ":" << msg << ": <" << cmd << ">" << endl;
	    ERR("create msg expected");
	  }

	  RT_GameParam param;

	  iss >> cpname >> cpind;
	  if (!iss) { cerr << msg << endl; ERR("create msg corrupt"); }

	  if (param.parse(cerr, iss)) ERR("parameters corrupt");
	  if (RTS_Game::check_param(cerr, param)) ERR("invalid parameters");
	  
	  opt.put("-type",   param.type);   // game type
	  opt.put("-x",      param.x);      // width
	  opt.put("-y",      param.y);      // height
	  opt.put("-pnum",   param.n);      // player number
	  opt.put("-speed",  param.s);      // speed
	  opt.put("-cpind",  cpind);        // client player index
	  opt.put("-cpname", cpname);       // client player name

	  cout << "client player index: " << cpind << endl;
	  cout << "client player name: " << cpname << endl;
	  
	  game = new RTS_Game(opt);
	  game->init();
	  state = ST_RECV_OBJS;
	  last_event = EVT_NONE;
	  
	  rt_client = new RT_Client(*this);
	  rt_client->run();

	  delete rt_client; rt_client = 0;
	  delete game; game = 0;

	  client_died = true; // -> initialize
	  cout << "game ended" << endl;
	  
	} else {

	  usleep(USLEEP);

	}
      }
      break;
      
    default: ERR("shouldn't get here");
    }
  }
}
