#ifndef BaseCommander_H
#define BaseCommander_H

/* 
 * Base Commander
 * Anthony Marcar
*/

#include "Commander.H"
#include "ClientModule.H"
#include "TerrainBase.H"

#include "Finance.H"

class GameStateModule;
class GameObj;
class BuildCommander;
class DefenseCommander;
class GatherCommander;

class BaseCommander : public Commander, ClientModule 
{
    struct BuildTask
    {
        std::string       bp_name;
        Finance::Ticket * funds;
        
        BuildTask(const std::string & bp_name,
                  Finance::Ticket * funds) :
            bp_name(bp_name), funds(funds) {}
    };

    class Building
    {
        ModuleSet &       ms;
        BaseCommander *   base;
        GameObj *         building;
        std::string       training;
        std::string       action;
        Finance::Ticket * funds;
        Watcher::WatchID  watch;
        bool              confirmed;
        bool              complete;
    public:
        Building(ModuleSet & ms, BaseCommander * base, GameObj * building);
        ~Building();
        GameObj * getBuilding();
        bool isBusy() const;
        const std::string & getTrainingBP() const;
        bool canTrain(std::string bp_name) const;
        void train(std::string bp_name, Finance::Ticket * funds);
        Watcher::WatchID getWatch() const;
        void update();
    };

    TerrainBase::Loc centreLoc;

    GameObj *          controlCenter;
    GatherCommander *  gatherer;
    DefenseCommander * defender;

    std::set<BuildCommander*> activeBuilders;
    std::set<BuildCommander*> finishedBuilders;

    std::map<std::string, size_t> unitsQueued; ///< Number of units queued
    std::map<GameObj*, Building*> obj2building; ///< Job information

    GameObj * builder;         ///< Unit set aside as a builder
    bool askingForBuilder;     ///< Whether or not we currently need builder

    std::queue<BuildTask> unitsToBuild;     ///< FIFO queue of training tasks
    std::queue<BuildTask> buildingsToBuild; ///< FIFO queue of build tasks


    //////////////////////////Methods////////////////////////////////

    ///initialises the Control Center and creates a gather commander
    void init(const Vector<GameObj*> & initialUnits);
    ///tells buildings to train any units in the queue
    void buildUnitsInQueue();

    void buildBuildingsInQueue();
    ///if there are any units in the unit vector, they are allocated to sub coms.
    void allocateExcessUnits();
    
    /**
     * called when this commander receives new units
     */
    void giveUnitNotification();

    /**
     * where everything is executed
     */
    void exec();

public:

    //*************************Constructor******************************************

    /**
     * creates a new base commander
     * @param mySuperior The commander controlling this one
     * @param units the units under this commanders control (includes buildings)
     * @param watcher a reference to the watcher
     * @param gsm a reference to the game state module
     * @param tm a reference to the terrainModule
     * @param st a reference to the pathfinding implementation
     * @param updateTime the amount of time between exec cycles
     */
    BaseCommander(Commander* const mySuperior, 
                  Vector<GameObj*> units, 
		  TerrainBase::Loc centre,
                  sint4 updateTime_=25);
    
    ~BaseCommander();

    void buildBuilding(const std::string & buildingBPname,
                       Finance::Ticket * funds);

    /**
     * adds units to be built to the queue
     * @param unitBPname the bp name of the unit to train
     */
    void buildUnit(const std::string & unitBPname,
                   Finance::Ticket * funds);

    /**
     * returns whether this commander can build the specified building
     * @param objectBPname the object to query (unit or building)
     * @return whether this unit can be built
     */
    bool canBuild(const std::string & objectBPname);

    TerrainBase::Loc getCentre() 
    {
        return centreLoc;
    }

    void onReceiveUnit( GameObj * unit, Watcher::WatchID );
    void onLoseUnit( GameObj * unit );

    /**
     * changes flags and sets vectors for the execution cycle.
     */
    bool onEvent(const Event &event);
    /**
     * returns whether this commander needs more units and asks sub commanders
     * @param unitBPname the bp name of the object to query
     * @return whether the unit is needed
     */
    bool needsUnit(const std::string & unitBPname);

    size_t countQueued(const std::string & bp_name) const;

    size_t countWorkersNeeded() const;

    virtual size_t countRequestable(const std::string & bp_name) const;

    virtual size_t requestUnits(Commander * receiver,
                                const std::string & bp_name,
                                size_t amount);
};

#endif
