#include "Commander.H"
    
/**
 * Decriments updateCounter by 1.
 * Recursively calls update on all Commanders.
 * When updateCounter == 0 updateCounter is set to 
 * a new random number (maxUpdateCount = maximum) and exec() is called.
 * @see updateCounter
 * @see maxUpdateCount
 * @see exec()
 */
void Commander::update() 
{
    if (updateCounter==0) 
    {
        exec();
        updateCounter=rand()%maxUpdateCount+1;
    }

    FORALL(comms, c)
    {
        (*c)->update();
    }

    updateCounter--;
}
    
/**
 * Constructor for Commander.
 * @param moduleSet reference to the global module set.
 * @param maxUpdate an integer that will be used as the highest possible random number of update() calls before exec() is run.
 * @see maxUpdateCount
 * @see updateCounter
 * @see ~Commander()
 */  
Commander::Commander(ModuleSet& moduleSet, int maxUpdate) :
    constructed(true),
    myCommander(NULL), 
    moduleSet(moduleSet), 
    maxUpdateCount(maxUpdate),
    myWatcher(moduleSet.get_watcher())
{ 
    srand((unsigned)time(0));
    updateCounter=lrand48()%maxUpdateCount+1;
    commanderID=EventFactory::new_who();
}

/**
 * Constructor for Commander.
 * @param mySuperior a pointer of type Commander to the Commander's superior.
 * @see myCommander
 * @param maxUpdate an integer that will be used as the highest possible random number of update() calls before exec() is run.
 * @see maxUpdateCount
 * @see updateCounter
 * @see ~Commander()
 */  
Commander::Commander(Commander* superior, int maxUpdate) :
    constructed(true),
    myCommander(superior), 
    moduleSet(superior->moduleSet), 
    maxUpdateCount(maxUpdate),
    myWatcher(moduleSet.get_watcher())
{ 
    srand((unsigned)time(0));
    updateCounter=lrand48()%maxUpdateCount+1;
    commanderID=EventFactory::new_who();

    superior->comms.insert(this);
}

/**
 * Destroys the Commander and all subcommanders recursively.
 * Sends units to superior Commander.
 * @see commVector
 * @see myCommander
 * @see unitVector
 * @see Commander()
 */
Commander::~Commander() 
{
    constructed = false;

    FORALL(comms, c)
    {
        delete (*c);
    }
    comms.clear();

    if(myCommander != NULL)
    {
        while( !units.empty() )
        {
            std::set<GameObj*>::iterator it = units.begin();
            GameObj * obj = (*it);
            removeUnit( obj );
            myCommander->addUnit( obj );
        }

        myCommander->comms.erase(this);
    }
}
    
bool Commander::handle_event(const Event &event)
{
    if( event.get_who() == Watcher::WatcherModule::FROM )
    {
        if( event.get_what() == Watcher::WatcherModule::CREATION )
        {
            const Watcher::CreationEvent & e =
                reinterpret_cast<const Watcher::CreationEvent&>(event);
            const std::set<GameObj*> nUnits = e.getUnits();
            std::set<GameObj*>::const_iterator it;
            for(it = nUnits.begin(); it != nUnits.end(); it++)
            {
                myWatcher.notifyOnDeath( this, (*it) );
                units.insert( (*it) );
                if( constructed )
                {
                    onReceiveUnit( (*it), e.getID() );
                }
            }
        }
        else if( event.get_what() == Watcher::WatcherModule::DEATH )
        {
            const Watcher::DeathEvent & e =
                reinterpret_cast<const Watcher::DeathEvent&>(event);
            const std::set<GameObj*> dUnits = e.getUnits();
            std::set<GameObj*>::const_iterator it;
            for(it = dUnits.begin(); it != dUnits.end(); it++)
            {
                if( constructed )
                {
                    onLoseUnit( (*it) );
                }
                units.erase( (*it) );
            }
        }
    }

    return onEvent(event);
}

sint4 Commander::getID() const 
{
    return commanderID;
}

void Commander::countObjects(const std::string & bp_name,
                             size_t & accumulator) const
{
    FORALL(units, u) 
    {
        if( (*u)->bp_name() == bp_name )
        {
            accumulator++;
        }
    }

    FORALL(comms, c)
    {
        (*c)->countObjects(bp_name, accumulator);
    }
}

void Commander::addUnit(GameObj * unit)
{
    myWatcher.notifyOnDeath( this, unit );
    units.insert( unit );
    if( constructed )
    {
        onReceiveUnit( unit, Watcher::NONE );
    }
}

void Commander::removeUnit(GameObj * unit)
{
    if( constructed )
    {
        onLoseUnit( unit );
    }
    units.erase( unit );
    myWatcher.cancelNotifyOnDeath( this, unit );
}

// Default implementations
size_t Commander::requestUnits(Commander *,
                               const std::string &,
                               size_t)
{
    return 0;
}

size_t Commander::countRequestable(const std::string &) const
{
    return 0;
}

// NOTE: demandUnits does not prioritize subcommanders in any
// significant way. If you want units to be stripped from commanders
// in a particular order, override requestUnits to specify how it
// should be done.
size_t Commander::demandUnits(Commander * receiver,
                              const std::string & bp_name,
                              size_t amount)
{
    size_t count = requestUnits(receiver, bp_name, amount);

    FORALL(comms, c)
    {
        if(count == amount)
        {
            break;
        }

        Commander * subCom = (*c);
        count += subCom->demandUnits(receiver, bp_name, amount - count);
    }

    std::set<GameObj*> toGive;
    FORALL(units, u)
    {
        if(count == amount)
        {
            break;
        }

        if( (*u)->bp_name() == bp_name )
        {
            toGive.insert( (*u) );
            count++;
        }
    }

    this->removeUnits(toGive);
    receiver->addUnits(toGive);

    return count;
}

size_t Commander::countOwned(const std::string & bp_name) const
{
    size_t count = 0;
    FORALL(comms, c)
    {
        Commander * subCom = (*c);
        count += subCom->countOwned(bp_name);
    }

    FORALL(units, u)
    {
        if( (*u)->bp_name() == bp_name )
        {
            count++;
        }
    }
    return count;
}
