// $Id: SimpleTerrain.C,v 1.1 2007/03/22 16:10:09 atelliot Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "SimpleTerrain.H"
#include "ST_PFEngine.H"
#include "Options.H"
#include "GameObj.H"
#include "Collision2D.H"
#include <utility>
#include <iterator>
#include <math.h>
#include "Profiler.H"
//#include "boost\smart_ptr.hpp"

using namespace std;
using boost::scoped_ptr;

#define TEST(x)
//#define TEST(x) x
#define TEST2(x)
//#define TEST2(x) x


namespace SimpleTerrain {

  const uint4 ST_Terrain::MAXIMUM_TASK_AGE = 250;          // milliseconds
  const sint4 ST_Terrain::ANTICIPATE_COLLISION_TICKS = 16; // 2 seconds * 8 ticks_per_second
  const sint4 ST_Terrain::FINE_GRID_RES_BITS = 6;
  const uint4 ST_Terrain::TILES_PER_MOTION_SECTOR = 3;

  //-----------------------------------------------------------------------------
  static Loc get_center_loc(const Object *obj)
  {
    Loc l;
    obj->get_center(l.x, l.y);
    return l;
  }

  static ScalarPoint get_center_sp(const Object *obj)
  {
    ScalarPoint l;
    obj->get_center(l.x, l.y);
    return l;
  }

  static real8 path_length(Path &path)
  {
    if( path.locs.size() < 2 )
      return 0;
    real8 total = 0;
    for( Vector<Loc>::iterator i = ++path.locs.begin() ; i != path.locs.end() ; i++ )
      total += i->distance(*(i-1));
    return total;
  }
    
  // Assuming path is in the usual reversed order, with goal at locs[0]
  static real8 path_length_from(Path &path, const Loc& start_loc)
  {
    if( path.locs.size() < 1 ) 
      return 0;
    return start_loc.distance(path.locs.back()) + path_length(path);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_options()
  {
    ::Options o("SimpleTerrain");
    o.put("-costonly", "represent boundaries only by higher edge cost");
    o.put("-highres", "use a high resolution world map (slower but more accurate)");
    o.put("-nomotionsectors", "don't use motion sectors for collision detection (test all pairs of units)");
  }

  //-----------------------------------------------------------------------------
  ST_Terrain::ST_Terrain() {
    sint4 seed = 0;
    ::Options::get("-seed", seed);
    rand.srand(seed);

    noMotionSectors = false;
    ::Options::get("-nomotionsectors", noMotionSectors);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::init(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, sint4 client_pID_, sint4 neutral_pID_)
  {
    TerrainBasicImp<ST_Task>::init(tiles_x_, tiles_y_, tile_points_, client_pID_, neutral_pID_);
    
    // initialize the internal PFEngine
    Vector<sint4> levels_of_detail;
    bool high_resolution_only;
    Options::get("-highres", high_resolution_only);
    if( !high_resolution_only ) {
      levels_of_detail.push_back(4);  // default - use a resolution reduced by a factor of 4.
    }
    levels_of_detail.push_back(1);  // also (or exclusively) maintain a full resolution view.
    pfEngine = boost::shared_ptr<PFEngine>(new PFEngine(tiles_x_, tiles_y_, tile_points_, levels_of_detail, client_pID));

    // initialize motion squares 
    //cout << "Created " << tiles_x_ * tiles_y_ << " sectors" << endl; 
/*    Vector<Sector> sectors_(tiles_x_ * tiles_y_);
    sectors = sectors_;
    sect_nx = tiles_x_;
  */  
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_obj(const Object *obj)
  {
    TerrainBasicImp<ST_Task>::add_obj(obj);
    assert(pfEngine!=NULL);
    pfEngine->insert_object(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::update_obj(const Object *obj)
  {
    TerrainBasicImp<ST_Task>::update_obj(obj);
    assert(pfEngine!=NULL);
    pfEngine->update_object(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::remove_obj(const Object *obj)
  {
    // shorthand
    typedef ObjIsGoalInTaskId::iterator OGIT;

    // remove from world representation
    assert(pfEngine!=NULL);
    pfEngine->remove_object(obj);

    // If we intended to stop it, we don't need to anymore
    objsToStop.erase(obj);

    // If this is the target of any existing task...
    pair<OGIT,OGIT> range = objIsGoalInTaskId.equal_range(obj);
    for( OGIT i = range.first ; i != range.second ; i++ ) {
      // Change the target of the task to the last known location of the object
      FIND(taskId2task, ti2tIt, i->second);
      assert(ti2tIt != taskId2task.end());
      ti2tIt->second->task.goal.target = Goal::LOCATION;
      ti2tIt->second->task.goal.loc = get_center_loc(ti2tIt->second->task.goal.obj);
      //ti2tIt->second->task.goal_obj->get_center(ti2tIt->second->task.goal_loc.x, 
      //                                          ti2tIt->second->task.goal_loc.y);
    }

    TerrainBasicImp<ST_Task>::remove_obj(obj);
  }

  //-----------------------------------------------------------------------------
  void ST_Terrain::add_segments(const Vector<Segment> &segs)
  {
    TerrainBasicImp<ST_Task>::add_segments(segs);
    assert(pfEngine!=NULL);
    FORALL(segs,i)
      pfEngine->insert_boundary(*i);
    invalidate_paths(segs);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::cancel_task(const Object *obj)
  {
    objsToStop.insert(obj);
    TerrainBasicImp<ST_Task>::cancel_task(obj);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::remove_task(TaskId tid)
  {
    FIND (taskId2task, it, tid);
    if (it != taskId2task.end() ) {
      Task& task = it->second->task;
      objsToStop.insert(task.objs.begin(), task.objs.end());
    }
    TerrainBasicImp<ST_Task>::remove_task(tid);
  }

  //Implemented in Forcefield;---------------------------------------------------------------------------
  real8 ST_Terrain::find_path(const Loc&, const Loc&, sint4, Path*, ConsiderObjects)
  {
    assert(0); // unimplemented yet    
    return -1;
  }

  //----------------------------------------------------------------------------------
  real8 ST_Terrain::find_path(const Object *obj, const Goal &goal, Path *path, ConsiderObjects consider)
  {
    assert( goal.mode == Goal::TOUCH && "ATTACK, VICINITY modes are currently unimplemented");
    assert( (goal.target == Goal::LOCATION || goal.target == Goal::OBJ) && 
            "Unknown Goal::target type");

    Path result;
    bool found;
    if (goal.target == Goal::LOCATION) {
      found = pfEngine->find_path(obj, goal.loc, result, 0, consider);
    } else {
      found = pfEngine->find_path(obj, goal.obj, result, 0, consider);
    }

    if( !found )
      return -1;

    // need to return waypoints in a 'forward' order
    if( path != NULL ) {
      path->locs.clear();
      reverse_copy(result.locs.begin(), result.locs.end(), back_inserter(path->locs));
    }

    // compute and return the length of the path
    return path_length_from(result, get_center_loc(obj));
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::block_path(const Object *obj, const Path &path)
  {
    assert(pfEngine!=NULL);
    pfEngine->ir_path(obj, path, true);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::release_path(const Object *obj, const Path &path)
  {
    assert(pfEngine!=NULL);
    pfEngine->ir_path(obj, path, false);
  }

  //----------------------------------------------------------------------------------
  TaskId ST_Terrain::execute_task(const Task &task, const Path &path)
  {
    TaskId tid = add_task(task);
    TIter ti = taskId2task[tid];
    // mark the task as not pending
    ti->isPending = false;
    eraseFromPending(ti);
    // and insert the requested path to execution
    Vector<Loc> reversed;
    reverse_copy(path.locs.begin(), path.locs.end(), back_inserter(reversed));
    ti->set_path(*task.objs.begin(), reversed);
    return tid;
  }

  //----------------------------------------------------------------------------------
  // fixme: we only check the path of the *center* of the unit vs the new boundaries,
  // and we should account for its size (radius).
  void ST_Terrain::invalidate_paths(const Vector<Segment> &segments)
  {
    FORALL(tasks, ti) {
      if (!ti->isValid() ) continue;
      // fixme: better handling of zcat != ON_LAND ?
      if (ti->path->obj->get_zcat() != Object::ON_LAND ) continue;

      FORALL(segments, si) {
        if (ti->path->is_intersect(*si) ) {
          // this path is invalid, mark as pending for (re)planning
          pendingTasks.push_back(ti);
          ti->isPending = true;
          cout << "Marked an invalid path" << endl;
          break;
        }
      }
    }
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::execute_tasks(Vector<MoveCmd> &cmds, Vector<StatusMsg> &msgs)
  {
    PROFILE("execute tasks");
    // cout << "ST_Terrain::execute_tasks" << endl;
    // Execute all active tasks
    /*for (Tasks::iterator j = tasks.begin(); j != tasks.end();) {
	Tasks::iterator i = j++;*/
    FORALL(tasks, i) {

      // Report failure of planning (no path at all)
      if (i->noPathFound ) {
        StatusMsg s;
        // fixme: this is not robust to future changes in the meaning of TaskId.
        s.task_id = (TaskId)&(*i);
        s.obj = *(i->task.objs.begin());
        s.type = StatusMsg::NO_PATH_FAILURE;
        msgs.push_back(s);
        continue;
      }

      // Skip invalid tasks.
      if (!i->isValid() ) { 
        TEST(cout << *i->task.objs.begin() << " task not valid" << endl;);
        continue;
      }

      UnitPath &cur_path = *(i->path);
      const Object *obj = cur_path.obj;

      // Skip tasks that are waiting for a retry
      if (cur_path.sleeping()) {
        TEST(cout << obj << " path sleeping" << endl;);
        continue;
      }

      // Skip objects for which we gave a (move) action that is still pending
      /*
      if (obj->is_pending_action()) {
      TEST(cout << obj << " pending action" << endl;)
      continue;
      }
      */

      bool atGoal = is_at_goal(obj, i->task);

      // Look ahead: if the object will get to a waypoint in the next tick,
      // give the next action now the prevent it from wasting a tick on stopping.
      // This will cut a single tick at each waypoint, though it will really prevent 
      // the object from stopping only if there's no (network or processing) lag
      // between this client and the server.
      // atWaypoint can be true only if the current waypoint is an intermediate and
      // not the *final* waypoint/goal.
      bool atWaypoint = cur_path.current_target > 0 &&  // reverse order, goal==0
        is_at_location(obj, cur_path.path.at(cur_path.current_target));

      // Skip a moving object which is currently not at a waypoint/goal
      if ((obj->get_speed() > 0 || obj->get_moving() != 0) && !atWaypoint && !atGoal ) {
        TEST(cout << obj << " atWaypoint=" << atWaypoint << " speed=" << obj->get_speed() << " is_moving=" << obj->get_moving() << endl;);
        continue;
      }

      // ok, so it needs attention:
      // 1. maybe it had reached the goal
      // 2. maybe it's at a waypoint (with the lookahead)
      // 3. maybe it needs the first orders
      // 4. maybe it unexpectedly collided with something

      if (atGoal ) {

        TEST(cout << obj << " arrived" << endl;)

        StatusMsg s;
        // fixme: this is not robust to future changes in the meaning of TaskId.
        s.task_id = (TaskId)&(*i);
        s.obj = obj; 
        s.type = StatusMsg::ARRIVED;
        msgs.push_back(s);

      } else if( atWaypoint || 
                 cur_path.no_orders_yet || 
                 cur_path.stopped_to_avoid_collision ) {

        // in all these cases we should just issue a move command

        TEST(cout << obj << " atWaypoint=" << atWaypoint
                  << " no_orders_yet=" << cur_path.no_orders_yet
                  << " stopped_to_avoid_collision=" << cur_path.stopped_to_avoid_collision
                  << endl; );

        // if we are at a waypoint, move to next waypoint
        if( atWaypoint )
          cur_path.current_target--;
        cur_path.no_orders_yet = false;
        cur_path.stopped_to_avoid_collision = false;
        add_move_command(cur_path, cmds);

      } else {

        // Unexpected stop. The current policy is:
        // 1-2. First time => wait a little and send the move command again.
        // 3-4. Second stop at the same place => mark as failed and replan.
        // 5. Third stop at the same place => report failure.

	TEST(cout << obj << " unexpectedly stopped, collision_step=" << cur_path.collision_step << endl);

        Loc cur_position = get_center_loc(obj);

        // First collision in this position:
        if (cur_path.spot != cur_position ) {
          cur_path.collision_step = 1;
          cur_path.mark_spot(cur_position);
          // sleep for a short random time, 100-599 millisecs
          cur_path.sleep(100 + (rand.rand_uint4() % 500));
        } else {
          cur_path.collision_step++;

          // Just finished the sleeping:
          if (cur_path.collision_step == 2 ) {
            add_move_command(cur_path, cmds);
          }
          // Second collision
          else if (cur_path.collision_step == 3 ) {
            i->isFailed = true;
          }
          // Back from replanning
          else if (cur_path.collision_step == 4 ) {
            add_move_command(cur_path, cmds);
          }
          // Third collision
          else if (cur_path.collision_step == 5 ) {
            StatusMsg s;
            // fixme: this is not robust to future changes in the meaning of TaskId.
            s.task_id = (TaskId)&(*i);
            s.obj = obj;
            s.type = StatusMsg::MOVEMENT_FAILURE;
            msgs.push_back(s);
	    TEST2(cout << "pf movement failure" << endl); 
          }
        }
      }
    }

    // Priority of commands:
    // 1. stops to avoid a collision.
    // 2. moves.
    // 3. other stops (resulting from a cancelled task, etc.).

    // 2 above 3:
    FORALL(cmds, c) {
      objsToStop.erase(c->obj);
    }

    // With all movement commands were given while assuming an ideal-no-collisions
    // world, now check if any of our controlled objects are going to collide
    // (because of an existing movement vector, and also consider commands
    // we just intended to give).
    resolve_friendly_collisions();

    // 1 above 2:
    for(typeof(cmds.begin()) c = cmds.begin() ; c != cmds.end() ; ) {
      FIND(objsToStop, s, c->obj);
      if( s != objsToStop.end() )
        c = cmds.erase(c);
      else
        c++;
    }

    // Return also the stop commands
    FORALL(objsToStop, oi) {
      const Object *obj = *oi;

      MoveCmd m;
      m.obj = obj;
      //obj->get_center(m.next_loc.x, m.next_loc.y); // doesn't really matter
      m.speed = 0; // 0 -> stop
      cmds.push_back(m);
    }
    objsToStop.clear();

    // remove finished tasks
    FORALL (msgs, t) {
      clean_task(t->task_id);
    }
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::add_move_command(const UnitPath& path, Vector<MoveCmd> &cmds)
  {
    MoveCmd m;
    m.obj = path.obj;
    m.next_loc = path.path.at(path.current_target);
    m.speed = path.obj->get_max_speed();
    cmds.push_back(m);
    TEST(cout << m.obj << " sending move command, target=(" << m.next_loc.x << "," << m.next_loc.y << ")" << endl;);
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::resolve_friendly_collisions(void)
  {
    // * Identify collisions between pairs of controlled units in the next few ticks.
    // * For each such expected collision, stop one of the objects in this turn (and
    //   then let the usual execution resume its movement later).
    // * While in the process of identifying pairs, remember the objects we already
    //   stopped and consider this in the following iterations.
    // * Assuming only one object per task

    scoped_ptr< SimpleMap<Sector> > sectors;
    if( !noMotionSectors ) {
      // To avoid the full O(N^2) comparisons, we insert objects into motion
      // sectors, and test only pairs that are in the same sector.
      
      // First, compute motion squares for all objects.
      Vector<UnitSquare> motionSquares;
      FORALL(tasks, t) {
        //  cout << "Computing motion square."<< endl;
        motionSquares.push_back(find_motion_square(t));
      }

      // Now, insert motion squares to all overlapping sectors.
      const sint4 motion_sectors_w = (tiles_x + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
      const sint4 motion_sectors_h = (tiles_y + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
      sectors.reset( new SimpleMap<Sector>(motion_sectors_w, motion_sectors_h) );
      insert_into_sectors(motionSquares, *sectors);
    }
    else {
      // Use a crude mechanism (check all-against-all), might be better for few units
      sectors.reset( new SimpleMap<Sector>(1,1) );
      FORALL(tasks, t)
        sectors->at(0).tasks.push_back(&*t);
    }

    // Walk the sectors, and compare pairs of tasks that are in the same sector.
    FORALL(*sectors, h) {

      // cout << "Checking " << active.size() << " sectors for collisions. " << endl;
      // cout << "This sector has " << (*h)->others.size() << " units in it. " << endl;

      // Test all pairs of objects/tasks:
      FORALL(h->tasks, i) {
        const Object *obj_i = *((*i)->task.objs.begin());

        typeof(i) j = i;
        for( j++ ; j != h->tasks.end() ; j++ ) {

          const Object *obj_j = *((*j)->task.objs.begin());
          //  cout << "Checking object: " << obj_i << " against object: " << obj_j << endl;

          if( will_collide(i, j) ) {
	    TEST(cout << "GOING TO COLLIDE: " << obj_i << "@" << get_center_loc(obj_i) << " and " << obj_j << "@" << get_center_loc(obj_j) << endl); 

            // current policy:
            // 1. try to (arbitrarily) stop the second object
            if( !will_collide(i, j, false, true) ) {
              // if stopping worked, j must have been moving, so we can 
              // safely use j->path
              objsToStop.insert(obj_j);
              (*j)->path->stopped_to_avoid_collision = true;
            }
            // 2. if the collision is still expected (eg they were heading 
            //    "into" one another), then stop both objects and 
            //    replan their paths
            else {
              // don't use the i->path, j->path since they might still not exist.
              objsToStop.insert(obj_i);
              objsToStop.insert(obj_j);
              (*i)->isFailed = true;
              (*j)->isFailed = true;
            }
          }
        }
      }
    }
  }

  //----------------------------------------------------------------------------------
  void ST_Terrain::get_fine_spatial_params(const Object *o, 
                                           ScalarPoint &target,
                                           ScalarPoint *fine_position, 
                                           ScalarPoint *fine_velocity)
  {
    ScalarPoint position = get_center_sp(o);
    ScalarPoint velocity = target;
    //o->get_center(position.x, position.y);
    //velocity = target;
    velocity.sub(position);

    // convert to fine-grid coordinates

    if( fine_position != NULL ) {
      fine_position->x = fine_grid(position.x);
      fine_position->y = fine_grid(position.y);
    }
    if( fine_velocity != NULL ) {
      fine_velocity->x =  fine_grid(velocity.x);
      fine_velocity->y =  fine_grid(velocity.y);
    }
    target.x = fine_grid(target.x);
    target.y = fine_grid(target.y);


    // normalize velocity to unit length and extend according to the object's speed

    if( fine_velocity != NULL && *fine_velocity != ScalarPoint(0,0) ) {
      // fixme: if object has a pending command, before moving, its speed is still 0. Should we use max_speed, or something else?
      if( o->get_speed() == 0 )
        *fine_velocity = ScalarPoint(0,0);
      else {
        real8 squared_norm = fine_velocity->l2_norm_sq();
        real8 inv_norm = 1.0 / sqrt(squared_norm);
        //cout << "squared_norm=" << squared_norm << " inv_norm=" << inv_norm << " fine_grid(o->get_speed())=" << fine_grid(o->get_speed()) << endl;
        fine_velocity->scale(inv_norm * fine_grid(o->get_speed()));
      }
    }
  }

  //----------------------------------------------------------------------------------
  bool ST_Terrain::will_collide(Vector<ST_Task*>::iterator i, Vector<ST_Task*>::iterator j, bool assume_obj1_stopped, bool assume_obj2_stopped)
  {
    // fixme: more efficent to cache obj1, since we test all pairs in a double loop

    // will these two objects collide in the next few ticks ?
    // look ahead only until the current waypoint for each object.
    // work with a higher accuracy.

    const Object *obj1 = *((*i)->task.objs.begin());
    ScalarPoint position1, velocity1, target1;
    real8 time_to_target1;

    // if it doesn't already have a move executed ==> immobile object
    if( !(*i)->isValid() || (*i)->path->sleeping() || (*i)->path->no_orders_yet || assume_obj1_stopped ) {
      // pass a dummy target and then adjust params
      ScalarPoint nn(0,0);
      get_fine_spatial_params(obj1, nn, &position1, NULL);
      velocity1 = ScalarPoint(0,0);
      time_to_target1 = ANTICIPATE_COLLISION_TICKS; // large enough to be ignored
    } 
    // else, the object is mobile
    else {
      target1 = ScalarPoint((*i)->path->path.at((*i)->path->current_target).x, (*i)->path->path.at((*i)->path->current_target).y);
      get_fine_spatial_params(obj1, target1, &position1, &velocity1);
      time_to_target1 = sqrt( (real8)(position1.l2_dist_sq(target1)) / velocity1.l2_norm_sq() );
    }


    const Object *obj2 = *((*j)->task.objs.begin());
    ScalarPoint position2, velocity2, target2;
    real8 time_to_target2;

    if( !(*j)->isValid() || (*j)->path->sleeping() || (*j)->path->no_orders_yet || assume_obj2_stopped ) {
      ScalarPoint nn(0,0);
      get_fine_spatial_params(obj2, nn, &position2, NULL);
      velocity2 = ScalarPoint(0,0);
      time_to_target2 = ANTICIPATE_COLLISION_TICKS;
    } else {
      target2 = ScalarPoint((*j)->path->path.at((*j)->path->current_target).x, (*j)->path->path.at((*j)->path->current_target).y);
      get_fine_spatial_params(obj2, target2, &position2, &velocity2);
      time_to_target2 = sqrt( (real8)(position2.l2_dist_sq(target2)) / velocity2.l2_norm_sq() );
    }

    // If they move at the same speed and at the same direction they won't collide
    if( velocity1 == velocity2 )
      return false;

    // if they will always to move like this, if and when will they collide?
    sint4 sum_of_radii = fine_grid(obj1->get_radius() + obj2->get_radius());
    real8 squared = (real8)(sum_of_radii) * sum_of_radii;

    /*James:  I found a bug.  When running this code for game2 with -nomotionsectors the following line produces an assert failure.
     It seems that when comparing two objects that are really far away the square function which is called in collision_time_circle_point()
     fails an assert.*/

    real8 collision_time = collision2d::collision_time_circle_point(position1, squared, velocity1, position2, velocity2);

    // will they collide at all, AND within the window of time that interests us,
    // AND before any of them reaches its target
    return collision_time != collision2d::no_collision && 
           collision_time <= ANTICIPATE_COLLISION_TICKS &&
           collision_time <= time_to_target1 &&
           collision_time <= time_to_target2;
  }

  //----------------------------------------------------------------------------------
  bool ST_Terrain::plan_tasks(uint4 max_time)
  {
    TEST(cout << "plan_tasks - max_time is : " << max_time << endl);
    uint4 stop_time = 0;
    if(max_time > 0){
      stop_time = SDL_GetTicks() + max_time;
    }
   
    // try to plan one failed and one pending tasks
    bool did_something = plan_failed_task();
    did_something = plan_pending_task() || did_something;

    // nothing more to plan at this time
    if( !did_something )
      return false;

    // Continue planning pending and failed alternately, 
    // until we have reached max_time or nothing more to plan
    bool plan_pending = true, did_pending = true, did_failed = true;
    while( (stop_time == 0 || SDL_GetTicks() < stop_time) && 
           (did_pending || did_failed) )
    {
      if( plan_pending )
	did_pending = plan_pending_task(stop_time);
      else
	did_failed = plan_failed_task(stop_time);
      plan_pending = !plan_pending;

      TEST(cout << "Current time is : " << SDL_GetTicks() << " and stop_time = " << stop_time << endl);
    }

#if 0
    bool did_some_more = plan_pending_task();
    sint4 counter = 0;
    //Continue planning until we have reached the max_time    
    while( (sint4)SDL_GetTicks() < stop_time && (!pendingTasks.empty() || !tasks.empty() )) {
      if(counter++ % 2 == 0)
	did_something = plan_failed_task( stop_time - (sint4)SDL_GetTicks() );
      else
	did_some_more = plan_pending_task( stop_time - (sint4)SDL_GetTicks() );

      TEST2(cout << "Current time is : " << (sint4)SDL_GetTicks() << " and stop_time = " << stop_time << endl);
 
      did_some_more = did_something || did_some_more; // careful not to short-circuit!
    }

    did_something = did_some_more || did_something; 
#endif
    return did_something;
  }

  //----------------------------------------------------------------------------------
  bool ST_Terrain::plan_pending_task(uint4 stop_time)
  {
    if (pendingTasks.empty()) return false;

    // TIter tit = pendingTasks.front(); pendingTasks.pop_front();

    std::deque<TIter>::iterator temp = get_best_task();
    TIter tit = *temp;
    pendingTasks.erase(temp);

    // mark the task as handled
    tit->isPending = false;

    // if the object is moving, stop it first and come back to it later
    assert(tit->task.objs.size() > 0);
    const Object *obj = *(tit->task.objs.begin());

    if (obj->get_speed() > 0 || obj->get_moving()) {
      objsToStop.insert(obj);
      pendingTasks.push_back(tit);
      return false;
      // fixme: might be better to try another pending task already in this tick...
    }

    // plan
    TEST(cout << obj << " planning pending path" << endl);
    plan_task(tit, stop_time);
    return true;
  }

  //-------------------------------------------------------------------
  struct FindFailed {
    bool operator()(ST_Task &task) { return task.isFailed && !task.noPathFound; }; 
  };

  //-------------------------------------------------------------------
  ST_Terrain::TIter ST_Terrain::get_random_failed_task(void)
  {
    // find a failed path for which we don't yet know that 'noPathFound'
    // (ie we still didn't try to replan and found that there is no path)
    if( tasks.empty() )
      return tasks.end();

    // start the search with in a random position to avoid planning the same
    // path repeatedly (in case it is immediately marked as failed again).
    int skip = rand(tasks.size());
    TIter start_at = tasks.begin();
    advance(start_at, skip);

    // search from the middle to the end
    TIter tit = find_if(start_at, tasks.end(), FindFailed());
    // if found
    if( tit != tasks.end() )
      return tit;
    // else, search from start to middle
    tit = find_if(tasks.begin(), start_at, FindFailed());
    return (tit != start_at)? tit : tasks.end();
  }

  //-------------------------------------------------------------------
  bool ST_Terrain::plan_failed_task(uint4 stop_time)
  {
    assert(pfEngine!=NULL);

    // get a failed task
    TIter tit = get_random_failed_task();
    if( tit == tasks.end() ) 
      return false;
  
    // if the object is moving, stop it first and come back to it later
    assert(tit->task.objs.size() > 0);
    const Object *obj = *(tit->task.objs.begin());

    if( obj->get_speed() > 0 || obj->get_moving() ) {
      objsToStop.insert(obj);
      return false;
      // fixme: might be better to try another failed task already in this tick...
    }

    // find-path
    TEST(cout << obj << " replanning path" << endl);
    plan_task(tit, stop_time);
    return true;
  }

  //-------------------------------------------------------------------
  void ST_Terrain::plan_task(TIter tit, uint4 stop_time)
  {
    PROFILE("plan a task");
    Path newpath;
    bool found;
    
    // Object to move
    const Object *obj = *(tit->task.objs.begin());

    if (tit->task.goal.target == Goal::LOCATION) {
      found = pfEngine->find_path(obj, tit->task.goal.loc, newpath, 0, CONSIDER_ALL, stop_time);
    } else {
      assert (tit->task.goal.target == Goal::OBJ);
      found = pfEngine->find_path(obj, tit->task.goal.obj, newpath, 0, CONSIDER_ALL, stop_time);
    }

    // insert results into data structures
    if( found ) {
      assert(newpath.locs.size() > 0);
      tit->set_path(obj, newpath.locs);
      tit->isFailed = false;
    }
    else {
      // fixme: maybe add to "failed"/sleep etc. ?
      tit->noPathFound = true;
    }

    // temp testing of ST_Terrain::find_path()
    /*Path temp;
    real8 len = find_path(obj, tit->task.goal, &temp);
    assert( equal(newpath.locs.begin(), newpath.locs.end(), temp.locs.rbegin())
            && newpath.locs.size() == temp.locs.size() && "Failed test" );*/

    // temp testing of ST_Terrain::block_path()
    /*block_path(obj, newpath);
    release_path(obj, newpath);*/
  }

  //-------------------------------------------------------------------

  bool ST_Terrain::is_at_location(const Object* obj, const Loc& target)
  {
    Loc position = get_center_loc(obj);
    return position.distance(target) <= obj->get_speed();
  }

  //-------------------------------------------------------------------
  
  bool ST_Terrain::is_near_location(const Object *obj, const Loc& target, sint4 range){
    Loc position = get_center_loc(obj);
    return position.distance(target) <= range;
  }

  //-------------------------------------------------------------------

  bool ST_Terrain::is_at_goal(const Object* obj, const Task& task)
  {
    typeof(obj2taskId.begin()) o2t;
    // Getting to the goal depends on the definition of the task:
    if (task.goal.target == Goal::OBJ ) {
      switch(task.goal.mode) {
      case Goal::TOUCH :
	return( obj->distance_to(*task.goal.obj) <= 1 );
	// using 1-tick look-ahead:
	//return( obj->distance_to(*task.goal_obj) <= obj->get_speed() );
	break;

      case Goal::VICINITY :
	
	return (obj->distance_to(*task.goal.obj) <= task.goal.distance);
	//assert(0 && "Not implemented yet");
	break;

      case Goal::ATTACK :

	return(is_in_range(obj, task.goal.obj));
             
	break;

      default:
	ERR("Unknown mode");
      }
    } 
    else if (task.goal.target == Goal::LOCATION ) {
      switch(task.goal.mode) {
      case Goal::TOUCH :
        // Temporary hack, until final approach is corrected for "LOCATION" 
        // targets: Since the planned path might finish at a location which is only
        // near the requested goal, compare to the planned goal instead of the 
        // requested goal. (When fixed, is_at_goal should be static again).
        //FIND(obj2taskId, o2t, obj);
        o2t = obj2taskId.find(obj);
        if( o2t != obj2taskId.end() ) {
          FIND(taskId2task, t2t, o2t->second);
          assert( t2t != taskId2task.end() );
          if( t2t->second->isValid() ) {
            return get_center_loc(obj) == t2t->second->path->path[0];
          }
        }
        // using 1-tick look-ahead; also suffers from low-resolution accuracy problems
	return is_at_location(obj, task.goal.loc);
	break;

      case Goal::VICINITY :
	return is_near_location(obj, task.goal.loc, task.goal.distance);
	//assert(0 && "Not implemented yet");
	break;

      case Goal::ATTACK :
	return(is_in_range(obj, task.goal.loc));
	//assert(0 && "Not implemented yet");
	break;

      default:
	ERR("Unknown mode");
      }
    } else {
      ERR("Unknown target type");
    }
    return false;
  }
  //-------------------------------------------------------------------

  bool ST_Terrain::is_in_range(const Object *moving_obj, const Loc goal_loc){
    //fixme: does not check if unit can see the target. 
    const GameObj *gob = dynamic_cast<const GameObj*>(moving_obj);
    sint4 range = 1;
    if (gob){
      ScriptObj *weapon = gob->component("weapon");
      const sintptr *r = weapon->get_int_ptr("max_ground_range");
      if (r){
	range = *r;
      }
    }

    Loc obj_loc = get_center_loc(moving_obj);
   
    return ( (obj_loc.distance(goal_loc) - moving_obj->get_radius()) <= range );

    
    // this is the check in ShooterAI. Other way makes more intuitive sense though...
    //     real8 rdist = moving_obj->distance_to(*goal_obj);
    //     sint4 sdist = (sint4)floor(rdist);
    //     return (sdist <= range);
  }

  //-------------------------------------------------------------------

  bool ST_Terrain::is_in_range(const Object *moving_obj, const Object *goal_obj){
    //fixme: does not check if unit can see each other. 
    const GameObj *gob = dynamic_cast<const GameObj*>(moving_obj);
    sint4 range = 1;
    if (gob){
      ScriptObj *weapon = gob->component("weapon");
      const sintptr *r = weapon->get_int_ptr("max_ground_range");
      if (r){
	range = *r;
      }
    }
  
    //if not a GameObj or with no weapon range, should behave exactly
    //like the check for touch.
    return ( moving_obj->distance_to(*goal_obj) <= range );
    
    // this is the check in ShooterAI. Other way makes more intuitive sense though...
    //     real8 rdist = moving_obj->distance_to(*goal_obj);
    //     sint4 sdist = (sint4)floor(rdist);
    //     return (sdist <= range);
  }

  //-------------------------------------------------------------------
  static bool is_younger(const TerrainBasicImp<ST_Task>::TIter &a, const TerrainBasicImp<ST_Task>::TIter &b) {
    return a->start_time > b->start_time;
  }

  static bool is_shorter(const TerrainBasicImp<ST_Task>::TIter &a, const TerrainBasicImp<ST_Task>::TIter &b) {
    return a->path_dist < b->path_dist; 
  }

  std::deque<ST_Terrain::TIter>::iterator ST_Terrain::get_best_task(void)
  {    
    // Check if the oldest task must be planned now
    std::deque<TIter>::iterator oldest = max_element(pendingTasks.begin(), 
                                                     pendingTasks.end(),
                                                     is_younger);
    if( SDL_GetTicks() - (*oldest)->start_time >= MAXIMUM_TASK_AGE )
      return oldest;

    // else, return the 'shortest' (straight-line) path
    return min_element(pendingTasks.begin(), pendingTasks.end(), is_shorter);
  }

  //-------------------------------------------------------------------
  void ST_Terrain::insert_into_sectors(const Vector<UnitSquare> &motionSquares,
                                       SimpleMap<Sector> &sectors)
  {
    PROFILE("Insert squares into sectors");
    const sint4 motion_sectors_w = (tiles_x + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
    const sint4 motion_sectors_h = (tiles_y + (TILES_PER_MOTION_SECTOR-1)) / TILES_PER_MOTION_SECTOR;
    const sint4 x_points_per_sector = (tiles_x * tile_points + motion_sectors_w-1) / motion_sectors_w;
    const sint4 y_points_per_sector = (tiles_y * tile_points + motion_sectors_h-1) / motion_sectors_h;

    FORALL (motionSquares, sqIt) {

      // square world-coords translated to sector-coords
      sint4 x0 = sqIt->topLeft.x / x_points_per_sector;     //(x - r);
      sint4 x1 = sqIt->bottomRight.x / x_points_per_sector; //(x + r);
      sint4 y0 = sqIt->topLeft.y / y_points_per_sector;     //(y - r);
      sint4 y1 = sqIt->bottomRight.y / y_points_per_sector; //(y + r);

      sint4 mw = sectors.get_w();
      sint4 mh = sectors.get_h();

      if (x0 < 0) x0 = 0;
      if (y0 < 0) y0 = 0;
      if (x1 >= mw) x1 = mw-1;
      if (y1 >= mh) y1 = mh-1;

      // insert to each sector
      for (sint4 y = y0; y <= y1; ++y) {
        for (sint4 x = x0; x <= x1; ++x) {
          //          if( x < 0 || x >= sectors.get_w() || y < 0 || y >= sectors.get_h() ){
          //            cout << "x = " << x << " y = " << y << endl;
          //          }
	  sectors(x,y).tasks.push_back(sqIt->task);
        }
      }
    }
  }
  
  //-------------------------------------------------------------------
  UnitSquare ST_Terrain::find_motion_square(const TIter &task)
  {
    PROFILE("find motion square");
    // If the object doesn't have a valid path
    if( !task->isValid() ) {
      // it is not mobile so use its current position
      const Object *obj = *task->task.objs.begin();
      Loc position = get_center_loc(obj);
      return UnitSquare(position, position, obj, &*task);
    }
    // A valid path
    else {
      const UnitPath &cur_path = *(task->path);
      const Object *obj = cur_path.obj;
      sint4 curx, cury; 
      obj->get_center(curx,cury);  // Current location

      // Find the position at a close future time  
      Loc future; 
      sint4 waypoints = get_future_position(cur_path, ANTICIPATE_COLLISION_TICKS, future);

      // Create a bounding box, enlarged by the obj's radius.
      std::vector<sint4> x_coords;
      std::vector<sint4> y_coords;

      FORS(i, waypoints+1){
        x_coords.push_back((*(cur_path.path.end()-i-1)).x);
        y_coords.push_back((*(cur_path.path.end()-i-1)).y);
      }    

      x_coords.push_back(future.x);
      y_coords.push_back(future.y);
      x_coords.push_back(curx);
      y_coords.push_back(cury);

       
      // Create a square containing the object and its motion. Out-of-bounds coords are OK.
      sint4 rad = obj->get_radius(); // The object's radius
      sint4 x0, y0, x1, y1;
      x0 = *min_element(x_coords.begin(), x_coords.end()) - rad;
      y0 = *min_element(y_coords.begin(), y_coords.end()) - rad;
      x1 = *max_element(x_coords.begin(), x_coords.end()) + rad;
      y1 = *max_element(y_coords.begin(), y_coords.end()) + rad;

      // constarin to map boundaries
      x0 = max(x0, 0);
      y0 = max(y0, 0);
      x1 = min(x1, tiles_x*tile_points - 1);
      y1 = min(y1, tiles_y*tile_points - 1);


     /*       
      Loc topLeft(x0 , y0);
      Loc bottomRight(x1 , y1);
      
      //  cout << "motion square is from (" << topLeft->x << "," << topLeft->y << ") to (" ;
      //   cout << bottomRight->x << "," << bottomRight->y << ")" << endl;
      */
      return UnitSquare(Loc(x0, y0), Loc(x1, y1), obj, &*task);
    }
  } 

  //-------------------------------------------------------------------
  sint4 ST_Terrain::get_future_position(const UnitPath &cur_path, sint4 when, Loc &where)
  {
    const Object *obj = cur_path.obj;
    where = get_center_loc(obj);
    
    sint4 v = obj->get_speed();
    if( v == 0 || when <= 0 || cur_path.path.size() == 0 ) {
      return -1;
    }

    real8 remaining_dist = v * when;
    sint4 got_to_waypoint = -1;
    real8 dist_to_next = 1; // =1 to prevent a 'might be used uninitialized' warning.
     
    TEST2(cout << "initially remaining_dist = " << remaining_dist << endl;);

    FORALL_REV_B(cur_path.path, next_waypoint_it) {

      dist_to_next = where.distance(*next_waypoint_it);
      remaining_dist -= dist_to_next;
      if(remaining_dist <= 0)
	break;
      where = *next_waypoint_it;
      got_to_waypoint++;
      TEST2(cout << "got to " << where << " with remaining_dist = " << remaining_dist << endl;);
    }
    
    // if not at end of path
    if( next_waypoint_it != cur_path.path.rend() ) {

      const Loc &next_waypoint = *next_waypoint_it;

      TEST2(cout << "current_pos = " << where << endl;
            cout << "next_waypoint = " << *next_waypoint_it << endl;
            cout << "dist_to_next = " << dist_to_next << endl;
            cout << "remaining_dist = " << remaining_dist << endl; );

      // compute the position inside the final leg
      if(remaining_dist == 0){
        where = next_waypoint;
        return got_to_waypoint+1;
      }

      real8 missing_to_next_wp_ratio = abs(remaining_dist/dist_to_next);
      assert(missing_to_next_wp_ratio != 0 );
      TEST2(cout << "missing_to_next_wp_ratio = " << missing_to_next_wp_ratio << endl;);
      
      sint4 dx = where.x - next_waypoint.x;
      sint4 dy = where.y - next_waypoint.y;

      where = Loc(next_waypoint.x + (sint4)ceil(dx*missing_to_next_wp_ratio - 0.5), 
                  next_waypoint.y + (sint4)ceil(dy*missing_to_next_wp_ratio - 0.5));
      //where = new Loc( next_waypoint.x + (sint4)ceil(x2*rat - 0.5) , next_waypoint.y + (sint4)ceil(y2*rat - 0.5));
    }
    TEST2(cout << "Object will be at " << where << " in " << when << " ticks." << endl;);
    return got_to_waypoint;
  }

  //----------------------------------------------------------------------------------
  //----------------------------------------------------------------------------------

#define EPS 1e-9f


  UnitPath::UnitPath(const Object *obj_, const Vector<Loc> &path_)
    : obj(obj_), path(path_), no_orders_yet(true), 
      stopped_to_avoid_collision(false), collision_step(0), wakeup(0)
  {
    current_target = path.size()-1;
    if (path.size() > 0) {
      goal = path[0];
    }
    spot = Loc(-1, -1);
  }


  bool UnitPath::is_intersect(const Segment &segment) const
  {
    typeof(path.begin()) pj = path.begin();
    typeof(path.begin()) pi = pj++;
    for( ; pj != path.end() ; pi++, pj++ ) {
      if (Segment(*pi, *pj).touches(segment) )
        return true;
    }
    return false;
  }


  //----------------------------------------------------------------------------------
  //----------------------------------------------------------------------------------

  ST_Task::ST_Task(const Task& task_)
    : task(task_), isPending(true), isFailed(false), noPathFound(false),
      start_time(SDL_GetTicks())
  {
    Loc start = get_center_loc(*task.objs.begin());
    if( task.goal.target == Goal::OBJ )
      path_dist = start.distance(get_center_loc(task.goal.obj));
    else if( task.goal.target == Goal::LOCATION )
      path_dist = start.distance(task.goal.loc);
    else
      ERR("Unknown Task::target value");
  }

  void ST_Task::set_path(const Object *obj_, const Vector<Loc> &path_)
  {
    // overwrite an existing path, retaining the collision-monitoring data.
    if( path ) {
      Loc   saved_spot = path->spot;
      int   saved_collision_step = path->collision_step;
      bool  saved_stopped_to_avoid_collision = path->stopped_to_avoid_collision;
      uint4 saved_wakeup = path->wakeup;

      path = boost::shared_ptr<UnitPath>(new UnitPath(obj_, path_));

      path->spot = saved_spot;
      path->collision_step = saved_collision_step;
      path->stopped_to_avoid_collision = saved_stopped_to_avoid_collision;
      path->wakeup = saved_wakeup;
    } 
    // No previous path was planned, so just create a new one.
    else {
       path = boost::shared_ptr<UnitPath>(new UnitPath(obj_, path_));
    }
  }

}   // End of namespace SimpleTerrain
