/** @file ClusterSquad.C
    @see ClusterSquad.H

    $Id: ClusterSquad.C,v 1.2 2007/05/25 01:16:45 sorsten Exp $
    $Source: /usr/bodo1/cvs/cvssoftware/orts3loc-2007/libs/squad2/src/ClusterSquad.C,v $
*/

// This is an ORTS file
// (c) Tim Furtak
// licensed under the GPL

#include "Global.H"
#include "ClusterSquad.H"
#include "GameObj.H"

using namespace std;

//===================================================================

const uint4 ClusterSquad::MAX_K = 20;

//===================================================================

ClusterSquad::ClusterSquad(sint4 x, sint4 y) : max_x(x), max_y(y)
{
  compute_baseline();
  prev_means.resize(MAX_K+1);
}

//===================================================================

ClusterSquad::~ClusterSquad()
{
}

//===================================================================

/** Squared distance. */

real4 ClusterSquad::distance(const Point &a, const Point &b) const
{
  real4 dx = abs(a.x - b.x);
  real4 dy = abs(a.y - b.y);
  real4 d = dx*dx + dy*dy;
  //  real4 d = dx + dy;
  return d;
}

//===================================================================

/** Computes the centroid of a group of points. */

void ClusterSquad::compute_mean(const Vector<Point> &p, Point &mean) const
{
  assert (p.size());

  mean.x = mean.y = 0;

  FORALL (p, it) {
    const Point &v = *it;
    mean.x += v.x;
    mean.y += v.y;
  }

  if (p.size()) {
    mean.x /= p.size();
    mean.y /= p.size();
  }
}

//===================================================================

/** Given a target point, finds the index of the element in the vector that is closest.
    Also computes the distance (err) to that element.
*/

sint4 ClusterSquad::closest(const Point &p, const Vector<Point> &means, real4 &err) const
{
  if (means.size() == 0) return -1;

  sint4 ind = 0;
  real4 d = distance(p, means[0]);

  for (uint4 i = 1; i < means.size(); ++i) {
    real4 e = distance(p, means[i]);
    if (e < d) {
      d = e;
      ind = i;
    }
  }

  err = d;
  return ind;
}

//===================================================================

/** Finds mean closest to a point p. */

sint4 ClusterSquad::classify(const Point &p) const
{
  real4 e;
  return closest(p, kmeans, e);
}

//-------------------------------------------------------------------

/** Finds mean closest to a GameObj g. */

sint4 ClusterSquad::classify(const GameObj *g) const
{
  Point p(*g->sod.x, *g->sod.y);
  return classify(p);
}

//===================================================================

sint4 ClusterSquad::classify(const GameObj *g, const ObjData &d)
{
  Point p(*g->sod.x, *g->sod.y);

  if (kmeans.size() == 0) return -1;

  sint4 ind = d.cluster_id;
  if (ind < 0) ind = 0;
  if (ind >= (sint4)kmeans.size()) ind = 0;

  real4 min_dist = distance(p, kmeans[ind]);

  FORS (i, kmeans.size()) {
    if (i == ind) continue;
    real4 e = distance(p, kmeans[i]);
    if (e < (min_dist - d.inertia)) {
      //      cout << "dist = " << e << endl;
      min_dist = e;
      ind = i;
    }
  }
  return ind;
}

//===================================================================

/** Reorders a vector of points so that their positions most closely match
    the positions of another vector - namely so that a new cluster assignment
    doesn't swap two means.
    Vector sizes don't have to match.
*/

void ClusterSquad::match_assignment(const Vector<Point> &orig, Vector<Point> &pts) const
{
  // greedy

  real4 err, errsum = 0;
  Vector<Point> pts2;

  uint4 min = ::min(orig.size(), pts.size());

  FORU (i, min) {
    sint4 ind = closest(orig[i], pts, err);
    errsum += err;
    if (ind < 0) ERR ("buh?");
    pts2.push_back(pts[ind]);
    pts[ind] = pts[pts.size()-1];
    pts.resize(pts.size()-1);
  }

  FORU (i, pts.size()) {
    pts2.push_back(pts[i]);
  }

  pts = pts2;
}


//===================================================================



//===================================================================

/** Compute baseline within-cluster classification error using
    1 to MAX_K means and points uniformly distributed across
    the game field.
*/

void ClusterSquad::compute_baseline()
{
  baseline.clear();
  baseline.resize(MAX_K+1);

  // number of sample points to scatter across the board
  const sint4 num_points = max_x * max_y / 4;

  Vector<real4> avg_e (MAX_K+1, 0.0); // average error
  Vector<real4> avg_e2(MAX_K+1, 0.0); // average squared error

  Vector<Point> means; // only used as a placeholder for 'compute_means'

  const uint4 trials = 1;

  cout << "baseline start ... " << flush;

  FORU (i, trials) {

    // scatter num_points in the unit cube
    Vector<Point> data;
    FORU (j, num_points) {
      Point p(lrand48() % max_x, lrand48() % max_y);
      data.push_back(p);
    }

    for (uint4 k = 1; k < MAX_K; ++k) {
      real4 err = log(compute_means(data, k, means));

      avg_e[k] += err;
      avg_e2[k] += err*err;
    }
  }

  cout << "done" << endl;

  real4 inv_t = 1.0 / trials;
  for (uint4 k = 1; k < MAX_K; ++k) {
    real4 avg_err = avg_e[k] * inv_t;
    real4 std_err = sqrt(avg_e2[k] * inv_t - avg_err*avg_err);
    UniformData ud;
    ud.err = avg_err;
    ud.std = std_err;
    baseline[k] = ud;
  }
}

//===================================================================

/** Compute a new set of means based on objects.
    Tries to avoid large changes in the reordering of existing means.
    Uses the gap heuristic described in "The Elements of Statistical Learning"
*/

void ClusterSquad::update_means()
{
  Vector< Vector<Point> > best_means;
  Vector<real4> min_err;
  Vector<real4> mu;
  Vector<real4> sigma;
  Vector<real4> gap;

  Vector<Point> means;

  const uint4 trials = 30;

  uint4 low =  1; //kmeans.size() - 2;
  uint4 high = kmeans.size() + 4;
  if (low < 1) low = 1;
  if (high > MAX_K) high = MAX_K;

  //  cout << "l,h = " << low << "  " << high << endl;

  real4 last_gap = -1e9;
  real4 cur_gap = last_gap;
  real4 my_baseline = 42;
  
  cout << "--------" << endl;

  uint4 best_k = 0;
  bool upswing = true;

  uint4 k;
  for (k = low; k <= high; ++k) {

    //    mu.push_back(baseline[k].err);    // average within-class classification error
    //    sigma.push_back(baseline[k].std); // standard deviation

    // compute best k-means error over some number of trials

    real4 err, b_err = 1e9;
    Vector<Point> bmeans;

    {
      means = prev_means[k];
      b_err = log(compute_means(kdata, k, means));
      bmeans = means;
    }

    /*
    if (k >= (sint4)kmeans.size()) {
      means = kmeans;
      b_err = log(compute_means(kdata, k, means));
      bmeans = means;
    }
    */

    FORU (i, trials) {
      means.clear();
      err = log(compute_means(kdata, k, means));
      //      cout << "log err = " << err << endl;
      if (err < b_err) {
        bmeans = means;
        b_err = err;
      }
    }
    best_means.push_back(bmeans);
    prev_means[k] = bmeans;

    min_err.push_back(b_err);

    real4 g = baseline[k].err - b_err; // gap for this k value (high is better)
    gap.push_back(g);
    cur_gap = g;
    my_baseline = baseline[k].std;

    cout << "g, berr, bl, last_gap = " << cur_gap << " " << b_err << " " << my_baseline << " " << last_gap << endl;

    real4 foo = 0.0;
    if (best_k) foo = ((k - best_k) - 1) * 0.15;

    // note that g is negative

    if ((1.0 + 0.0*k)*g - baseline[k].std < last_gap + foo) {
      if (upswing) {
        //        best_k = k - 1;
        upswing = false;
      }
    } else {
      best_k = k;
      last_gap = g;
      upswing = true;
    }

  }
  if (best_k == 0) {
    k = high;
  } else {
    k = best_k;
  }
  if (k < low) k = low;
  if (k > high) k = high;
  
  if (kmeans.size() != k) {
    cout << "switching " << kmeans.size() << " -> " << k << "  " << endl;
  }

  match_assignment(kmeans, best_means[k-low]); // reorder the best means to better match the old numbering

  kmeans = best_means[k-low];
  //  prev_means[k] = kmeans;
  /*
  cout << "GAP:" << endl;
  printf("K\tmu\tsigma\terr\tgap\n");
  FORU (i, gap.size()) {
    printf("%d\t%.4f\t%.4f\t%.4f\t%.4f\n", int(low+i), mu[i], sigma[i], min_err[i], gap[i]);
  }
  cout << "k = " << k << "  classification error: " << min_err[k-low] << endl;
  */
}

//===================================================================

/** Randomly selects k means, and iterates until a stable solution is reached
    where each point is assinged to one mean.

    @return Computed classification error.
*/

real4 ClusterSquad::compute_means(const Vector<Point> &data, sint4 k, Vector<Point> &means) const
{
  if (k < (sint4)means.size()) means.clear();
  assert(data.size());
  assert(k);

  for (sint4 i = means.size(); i < k; ++i) {
    sint4 p = lrand48() % data.size();
    means.push_back(data[p]);
  }

  Vector<Point> old_means = means;

  real4 err;

  bool diff = true;

  Vector< Vector<sint4> > bin;

  // iterate while the mean is still changing ==> group assignments are still changing

  sint4 c = 0;

  while (diff && c < 100) {
    ++c;

    assign_hard(data, means, bin, err); // computes the new group assignment and also some
                                        // error measure for the assignment
    
    FORU (i, bin.size()) {
      Vector<Point> group;
      FORALL (bin[i], it) {
        group.push_back(data[*it]);
      }
      if (group.size()) {
        compute_mean(group, means[i]); // recompute the mean from the average of all elements assigned to that group
      }
    }

    diff = false;
    FORU (i, k) {
      assert (i < means.size());
      assert (i < old_means.size());
      const Point &m1 = means[i];
      const Point &m2 = old_means[i];

      if ((m1.x != m2.x) || (m1.y != m2.y)) {
        diff = true;
        break;
      }
    }
    old_means = means;
  }

  return err;
}

//===================================================================

/** Classifies a vector of points, placing the id of each point in the assignment
    vector that corresponds to the appropriate bin.
*/

void ClusterSquad::assign_hard(const Vector<Point> &p, const Vector<Point> &means,
                               Vector< Vector<sint4> > &bin, real4 &err) const
{
  err = 0.0;
  real4 e = 0.0;

  bin.resize(means.size());
  FORALL (bin, it) it->clear();

  FORU (i, p.size()) {
    sint4 id = closest(p[i], means, e);
    bin[id].push_back(i);
    err += e;
  }
}

//===================================================================

void ClusterSquad::add(const GameObj *obj)
{
  ObjData d;
  d.cluster_id = -1;
  d.inertia = 1024; // (2*16)^2
  obj2dat[obj] = d;
}

//-------------------------------------------------------------------

void ClusterSquad::del(const GameObj *obj)
{
  obj2dat.erase(obj);
}

//===================================================================

void ClusterSquad::update()
{
  kdata.clear();
  FORALL (obj2dat, it) {
    const GameObj *g = it->first;
    Point p;
    p.x = *g->sod.x;
    p.y = *g->sod.y;
    kdata.push_back(p);
  }

  update_means();

  FORALL (obj2dat, it) {
    const GameObj *g = it->first;
    ObjData &d = it->second;

    sint4 nc = classify(g, d);
    d.cluster_id = nc;
  }
}

//===================================================================
