#ifndef SquadCombatAI_H
#define SquadCombatAI_H

// $Id: SquadCombatAI.H,v 1.3 2007/05/25 01:16:45 sorsten Exp $

// This is an ORTS file (c) Tim Furtak, licensed under the GPL

#include "Global.H"
#include "EventHandler.H"
#include "ClientModule.H"
#include "InfluenceMap.H"
#include "VectorField.H"
#include "TerrainVectorField.H"
#include "GameObj.H"

class GameStateModule;
class GameChanges;

//===================================================================

class SquadCombatAI : public EventHandler, public ClientModule
{
public:

  static void add_options();

  //------------------------------

  /** Top-level collection of related (allied) objects. */
  // Public only to allow REGISTER_TYPEOF definitions, needed e.g. in MSVC .NET 2003
  class GroupData {
  public:
    std::set<GameObj::WPtr> units;
    InfluenceMap attack_influence;
    InfluenceMap target_influence;
    InfluenceMap occupied_influence;

    VectorField attack_field;
    VectorField target_field;
    VectorField occupied_field;
    VectorField motion_field;

    GroupData(sint4 width, sint4 height);
  };

  //------------------------------

protected:

  //------------------------------

  /** Units with a common objective. */

  class Squad {
  public:

    enum Objective { ATTACK_TARGET, ATTACK_LOCATION, FREELANCE, MOVE_TO_POINT };

    std::set<GameObj::WPtr> units;
    sint4 x, y;
    GameObj::WPtr target;
    Objective objective;

    real4 pos_x, pos_y;

    real4 avg_dist_to_enemy;
    real4 max_diff_to_enemy;

    Squad() : x(0), y(0), target(0), objective(ATTACK_TARGET) {}
  };

  //------------------------------

  class ShooterData {
  public:
    sint4 x, y, r;
    sint4 min_damage;
    sint4 max_damage;
    sint4 avg_damage;
    sint4 range, air_range;
    sint4 hp;
    sint4 armor;
    sint4 time_to_shoot;
    GameObj::ZCat zcat;
    bool line_of_sight; ///< true if unit needs to see the enemy to shoot it
    Vector<GameObj::WPtr> outgoing; ///< potential targets
    Vector<GameObj::WPtr> incoming; ///< potential threats
    GameObj::WPtr cur_target;
    GameObj::WPtr old_target;

    Vec2<real4> dir; ///< Intended direction ignoring negative effects.
    Vec2<real4> ndir; ///< Normalized.

    Vec2<real4> force1, force2; // attack all, attack target
    Vec2<real4> to_centroid;

    real4 support; ///< Measure of nearby friendly units.
    real4 dist_to_enemy; ///< Distance to the nearest enemy.
    real4 diff_to_enemy;

    bool motion_enabled;

    enum Mode { ACTIVE, PASSIVE };
    Mode mode;

    Squad *squad;

    ShooterData();
  };

  //------------------------------
public:
  // Public only to allow REGISTER_TYPEOF definitions, needed e.g. in MSVC .NET 2003
  typedef std::map<GameObj*, ShooterData> ShooterMap;

  //------------------------------

protected:

  GameStateModule &gsm;
  uint4 game_tick;     // game time
  sint4 width, height; // map size in tile points

  Vector<GroupData*> groups;
  GroupData *friendlies;
  GroupData *enemies;
  GroupData *neutrals;

  ShooterMap shooter_data;
  Vector<Squad*> squads;

  TerrainVectorField terrain_field;
  VectorField occupied_field;

  void update_units();
  void remove_obj(const GameObj::WPtr gob);

  GroupData* get_group(GameObj::WPtr gob);


  void compute_influence();
  void compute_influence(GroupData *gd);
  void compute_occupation(GroupData *gd);

  void compute_attack_graph();
  void compute_attack_graph(GroupData *ff, GroupData *ee);

  void compute_support(const std::set<GameObj::WPtr> &units);
  void compute_support2(const std::set<GameObj::WPtr> &units);

  real4 support(const Vec2<real4> &pos, const Vec2<real4> &dir,
                const Vec2<real4> &spos, const Vec2<real4> &sdir);
  real4 support(const Vec2<real4> &pos, const Vec2<real4> &spos);


  Vec2<real4> attack_vec(sint4 x, sint4 y, real4 dx, real4 dy);


  Vec2<real4> pro_vec(sint4 x, sint4 y, real4 dx, real4 dy);
  Vec2<real4> spacing_vec(sint4 x, sint4 y, real4 dx, real4 dy);
  Vec2<real4> anti_vec(sint4 x, sint4 y, real4 dx, real4 dy);
  
  void add_to_squad(GameObj::WPtr g, Squad *s, ShooterData::Mode m=ShooterData::ACTIVE);
  void update_squads();

  bool enabled_motion;
  bool enabled_attack;

  //  void compute_motion();
  void compute_attack();

  void compute_squad_motion2();

  //-------------------------------------
  // Shooter AI
  //-------------------------------------

  void add_friendly(GameObj*);
  void build_range_graph();
  void removeNode(GameObj *node);
  void recompute_targets();
  
  real8 unitRank(GameObj *enemy);
  real8 unitRandom(/*GameObj *enemy*/);
  sint4 unitHealth(GameObj *enemy);
  sint4 unitAttack(GameObj *unit);
  real8 unitImportance(GameObj *enemy);
  real8 avgEnemyImportance(GameObj *enemy);

  bool cmpAttack(GameObj *a, GameObj *b);
  bool cmpEnemyImportance(GameObj *a, GameObj *b);
  void chooseAttackers(Vector<GameObj*> &attackers, GameObj *unit,
                       const Vector<GameObj::WPtr> &possible_attackers);

  static const real8 overkill_percent = 0.0;


  enum Strategy { LOCAL_STRATEGY=1, GLOBAL_STRATEGY };
  Strategy strategy;

  enum TargetSelectionStrategy { SHOOT_NONE=0, SHOOT_RANDOM, SHOOT_WEAKEST,
                                 SHOOT_IMPORTANT, SHOOT_IMPORTANT_2 };
  TargetSelectionStrategy target_strategy;


  //-------------------------------------

public:

  static const sint4 FROM;
  static const sint4 SET_PRIORITY;
  static const sint4 ATTACK_TARGET;
  static const sint4 ATTACK_LOCATION;
  static const sint4 ATTACK_FREELANCE;
  static const sint4 MOVE_LOCATION;
  static const sint4 STOP_ATTACK;

  static const sint4 ENEMY_IN_RANGE;
  static const sint4 TARGET_DESTROYED;
  static const sint4 AT_LOCATION;
  //  static const sint4 TARGET_SECURED;

  SquadCombatAI(GameStateModule &gsm);
  ~SquadCombatAI();

  bool handle_event(const Event &);

  void update(const GameChanges &changes);

  void enable_attack() { enabled_attack = true; }
  void enable_motion() { enabled_motion = true; }
  void enable_unit_motion(GameObj * unit)
  {
      shooter_data[unit].motion_enabled = true;
  }

  void disable_attack() { enabled_attack = false; }
  void disable_motion() { enabled_motion = false; }
  void disable_unit_motion(GameObj * unit)
  {
      shooter_data[unit].motion_enabled = false;
  }

  bool is_enabled_attack() const { return enabled_attack; }
  bool is_enabled_motion() const { return enabled_motion; }

  friend class SquadCombatAIWidget;
  friend class SquadCombatAIOverlay;
};

//===================================================================
//===================================================================

class ShooterPriorityEvent : public Event
{
public:
  
  enum Priority { ACTIVE, PASSIVE };

  ShooterPriorityEvent(sint4 who, sint4 what, GameObj::WPtr o, Priority p) :
    Event(who, what), obj(o), priority(p) {}

  ~ShooterPriorityEvent() {}
  
  GameObj::WPtr obj;
  Priority priority;
};

//===================================================================

class SquadCombatEvent : public Event
{
public:
  
  enum Objective { STOP, ATTACK_TARGET, ATTACK_LOCATION }; // MOVE_TO_POINT

  SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj::WPtr> &u, sint4 x, sint4 y);
  SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj::WPtr> &u, GameObj::WPtr targ);
  SquadCombatEvent(sint4 who, sint4 what, const Vector<GameObj::WPtr> &u);
  ~SquadCombatEvent() {}
  
  Vector<GameObj::WPtr> units;
  sint4 x, y;
  GameObj::WPtr target;
  Objective objective;
};

//===================================================================

REGISTER_TYPEOF(2881, Vector<SquadCombatAI::GroupData*>::iterator);
REGISTER_TYPEOF(2882, Vector<SquadCombatAI::GroupData*>::const_iterator);
REGISTER_TYPEOF(2883, SquadCombatAI::ShooterMap::iterator);
REGISTER_TYPEOF(2884, SquadCombatAI::ShooterMap::const_iterator);
REGISTER_TYPEOF(2885, Vector<SquadCombatAI::Squad*>::iterator);
REGISTER_TYPEOF(2886, Vector<SquadCombatAI::Squad*>::const_iterator);


#endif
