#include "BaseClassifier.H"
#include "WatcherAux.H"

namespace TerrainAnalysis
{
    BaseInfo::BaseInfo(const ResourceClusterInfo & resources) :
        resources(resources)
    {
        
    }

    void BaseInfo::addAlliedBuilding(const History::Record * building)
    {
        alliedBuildings.insert(building);
    }

    void BaseInfo::addEnemyBuilding(const History::Record * building)
    {
        enemyBuildings.insert(building);
    }

    size_t BaseInfo::countAlliedBuildings() const
    {
        return alliedBuildings.size();
    }

    size_t BaseInfo::countAlliedBuildings(const std::string & bp_name) const
    {
        size_t size = 0;
        FORALL(alliedBuildings, b)
        {
            if( (*b)->isActive() )
            {
                if( (*b)->getObject()->bp_name() == bp_name )
                {
                    size++;
                }
            }
        }
        return size;
    }

    size_t BaseInfo::countEnemyBuildings() const
    {
        return enemyBuildings.size();
    }

    size_t BaseInfo::countEnemyBuildings(const std::string & bp_name) const
    {
        size_t size = 0;
        FORALL(enemyBuildings, b)
        {
            if( (*b)->isActive() )
            {
                if( (*b)->getObject()->bp_name() == bp_name )
                {
                    size++;
                }
            }
        }
        return size;
    }

    TerrainBase::Loc BaseInfo::getResourceLocation() const
    {
        return resources.location;
    }

    TerrainBase::Loc BaseInfo::getCentralLocation() const
    {
        TerrainBase::Loc loc = resources.location;
        sint4 count = 1;
        FORALL(alliedBuildings, b)
        {
            if( (*b)->isActive() )
            {
                const GameObj * obj = (*b)->getObject();
                sint4 x, y; obj->get_center(x, y);
                loc.x += x; loc.y += y;
                count++;
            }
        }
        FORALL(enemyBuildings, b)
        {
            if( (*b)->isActive() )
            {
                const GameObj * obj = (*b)->getObject();
                sint4 x, y; obj->get_center(x, y);
                loc.x += x; loc.y += y;
                count++;
            }
        }
        loc.x /= count;
        loc.y /= count;
        return loc;
    }

    sint4 BaseInfo::resourceWorth() const
    {
        return resources.estimatedWorth;
    }

    sint4 BaseInfo::alliedWorth() const
    {
        sint4 worth = 0;
        FORALL(alliedBuildings, b)
        {
            if( (*b)->isActive() )
            {
                const GameObj * obj = (*b)->getObject();
                const sintptr * cost = obj->get_int_ptr("cost_mineral");
                if(cost != NULL)
                {
                    worth += (*cost);
                }
            }
        }
        return worth;
    }

    sint4 BaseInfo::enemyWorth() const
    {
        sint4 worth = 0;
        FORALL(enemyBuildings, b)
        {
            if( (*b)->isActive() )
            {
                const GameObj * obj = (*b)->getObject();
                const sintptr * cost = obj->get_int_ptr("cost_mineral");
                if(cost != NULL)
                {
                    worth += (*cost);
                }
            }
        }
        return worth;
    }

    BaseInfo::OWNER BaseInfo::getOwner() const
    {
        if(alliedBuildings.empty())
        {
            if(enemyBuildings.empty())
            {
                return UNCLAIMED;
            }
            else
            {
                return ENEMY;
            }
        }
        else
        {
            if(enemyBuildings.empty())
            {
                return ALLIED;
            }
            else
            {
                return CONTESTED;
            }
        }
    }

    BaseClassifier::BaseClassifier(ModuleSet & ms) : ms(ms)
    {

    }

    BaseClassifier::~BaseClassifier()
    {
        clear();
    }

    void BaseClassifier::clear()
    {
        FORALL(bases, b)
        {
            delete (*b);
        }
        bases.clear();
        const_bases.clear();
    }

    void BaseClassifier::update()
    {
        sint4 me = ms.get_game_state().get_game().get_client_player();

        Vector<ResourceClusterInfo> clusters;
        ms.get_terrain_analysis().find_resource_clusters("mineral", clusters);
        FORALL(clusters, c)
        {
            BaseInfo * base = new BaseInfo( (*c) );
            bases.push_back(base);
            const_bases.push_back(base);
        }

        Vector<const History::Record *> buildings;
        ms.get_history().queryActiveRecords( Watcher::IsBuilding(), 
                                             buildings );
        FORALL( buildings, r )
        {
            const History::Record * record = (*r);
            const GameObj * object = record->getObject();

            if( object->sod.owner == NULL ) continue;
            sint4 owner = *(object->sod.owner);
            TerrainBase::Loc objLoc; object->get_center(objLoc.x, objLoc.y);

            BaseInfo * best = NULL;
            real8 bestDistance = 0;
            FORALL( bases, b )
            {
                BaseInfo * base = (*b);
                TerrainBase::Loc loc = base->getResourceLocation();
                sint4 dx = objLoc.x - loc.x;
                sint4 dy = objLoc.y - loc.y;
                real8 distance = dx*dx + dy*dy;
                if( best == NULL || distance < bestDistance )
                {
                    best = base;
                    bestDistance = distance;
                }
            }

            if(best != NULL)
            {
                if( owner == me )
                {
                    best->addAlliedBuilding(record);
                }
                else
                {
                    best->addEnemyBuilding(record);
                }
            }
        }
    }
    
    const Vector<const BaseInfo*> & BaseClassifier::get_bases() const
    {
        return const_bases;
    }
}
