#include "TerrainAnalysis.H"
#include "WatcherAux.H"

namespace TerrainAnalysis 
{
    void FindBorderLocations(const Map<GameTile> & map, sint4 ppt,
                             Vector<TerrainBase::Loc> & results)
    {
        sint4 width = map.get_width();
        sint4 height = map.get_height();
        sint4 half = ppt >> 1;

        char * buffer = new char[width*height];
        char * data = buffer;
        for(sint4 y = 0; y < height; y++)
        {
            for(sint4 x = 0; x < width; x++)
            {
                const GameTile & tile = map.get_tile(x, y);
                Tile::Type tw, te; tile.get_type(tw, te);
                char type = 0;
                if(tw == Tile::UNKNOWN ||
                   te == Tile::UNKNOWN) type = 1;
                else if(tile.get_min_h() != 0 ||
                        tile.get_max_h() != 0) type = 2;
                *data++ = type;
            }
        }

        data = buffer + width; // Skip first row
        sint4 x0 = 1, x1 = width - 1;
        sint4 y0 = 1, y1 = height - 1;
        for(sint4 y = y0; y < y1; y++)
        {
            data++; // Skip first column
            for(sint4 x = x0; x < x1; x++)
            {
                if( *data == 1 ) // Unknown tile
                {
                    // If tile has known, zero height border tile
                    if( data[1] == 0 || data[-1] == 0 ||
                        data[width] == 0 || data[-width] == 0 )
                    {
                        sint4 px = x * ppt + half;
                        sint4 py = y * ppt + half;
                        results.push_back(TerrainBase::Loc(px,py));
                    }
                }
                data++; 
            }
            data++; // Skip last column
        }
    }

    TerrainRegion::TerrainRegion(sint4 x, sint4 y, 
                                 sint4 width, sint4 height, sint4 submaps) :
        x(x), y(y), width(width), height(height), submaps(submaps) 
    {
        br = tr = bl = tl = 0;
        unexplored = width * height;
    }

    void TerrainRegion::add_tile(const Loc &loc) 
    {
        assert(loc.x >= x && loc.x <= (x + width));
        assert(loc.y >= y && loc.y <= (y + height));

        unexplored--;

        sint4 lx = loc.x - x, ly = loc.y - y, w_2 = width / 2, h_2 = height / 2;

        if (!submaps) return; // Don't keep nesting if no more submaps
        
        if (lx > w_2)
        {
            if (ly > h_2) 
            {
                if (!br)
                {
                    br = new TerrainRegion(x + w_2, y + h_2, 
                                           w_2, h_2, submaps - 1);
                }
                br->add_tile(loc);
            } 
            else 
            {
                if (!tr)
                {
                    tr = new TerrainRegion(x + w_2, y, w_2, h_2, submaps - 1);
                }
                tr->add_tile(loc);
            }
        }
        else
        {
            if (ly > h_2) 
            {
                if (!bl)
                {
                    bl = new TerrainRegion(x, y + h_2, w_2, h_2, submaps - 1);
                }
                bl->add_tile(loc);
            }
	    else 
            {
                if (!tl)
                {
                    tl = new TerrainRegion(x, y, w_2, h_2, submaps - 1);
                }
                tl->add_tile(loc);
            }
	}
    }

    void TerrainRegion::find_largest_unexplored_area(Loc &loc) 
    {
        TerrainRegion *target = 0;

        if (!(tl || tr || bl || br) || submaps <= 0) 
        {
            loc.x = x + width / 2;
            loc.y = y + height / 2;
            
            return;
        }

        sint4 w_2 = width / 2, h_2 = height / 2;

        if (!tl) 
        {
            tl = target = new TerrainRegion(x, y, w_2, h_2, submaps - 1);
        }
        else if (!tr)
        {
            tr = target = new TerrainRegion(x + w_2, y, w_2, h_2, submaps - 1);
        }
        else if (!bl) 
        {
            bl = target = new TerrainRegion(x, y + h_2, w_2, h_2, submaps - 1);
        }
        else if (!br)
        {
            br = target = new TerrainRegion(x + w_2, y + h_2, 
                                            w_2, h_2, submaps - 1);
        }

        if (!target) 
        {
            target = tl;
            if (tr->get_unexplored() > target->get_unexplored()) target = tr;
            if (bl->get_unexplored() > target->get_unexplored()) target = bl;
            if (br->get_unexplored() > target->get_unexplored()) target = br;
        }

        target->find_largest_unexplored_area(loc);
    }

    TerrainRegion::~TerrainRegion() 
    {
        if (tl) delete tl;
        if (tr) delete tr;
        if (bl) delete bl;
        if (br) delete br;
    }

    class ResourceCluster
    {
        std::vector<const History::Record*> resources;
    public:
        ResourceCluster(const History::Record * first)
        {
            resources.push_back(first);
        }

        real8 minDistance(const History::Record * next) const
        {
            assert( next->isActive() );

            real8 minDistance = 2048.0;
            std::vector<const History::Record*>::const_iterator it;
            for(it = resources.begin(); it != resources.end(); it++)
            {
                if( (*it)->isActive() )
                {
                    const GameObj * obj0 = (*it)->getObject();
                    const GameObj * obj1 = next->getObject();
                    real8 distance = obj0->distance_to(*obj1);
                    if(distance < minDistance) minDistance = distance;
                }
            }
            return minDistance;
        }
        
        void addResource(const History::Record * next)
        {
            resources.push_back(next);
        }

        TerrainBase::Loc getCentroid() const
        {
            sint4 tx = 0, ty = 0;
            std::vector<const History::Record*>::const_iterator it;
            for(it = resources.begin(); it != resources.end(); it++)
            {
                if( (*it)->isActive() )
                {
                    sint4 x, y; (*it)->getObject()->get_center(x, y);
                    tx += x; ty += y;
                }
            }                
            size_t size = resources.size();
            return TerrainBase::Loc(tx / size, ty / size);
        }
        
        sint4 countPatches() const
        {
             return static_cast<sint4>(resources.size());
        }

        sint4 estimateWorth() const
        {
            sint4 total = 0;
            std::vector<const History::Record*>::const_iterator it;
            for(it = resources.begin(); it != resources.end(); it++)
            {
                if( (*it)->isActive() )
                {
                    const GameObj * obj = (*it)->getObject();
                    const sintptr * resource = obj->get_int_ptr("hp");
                    if( resource != NULL )
                    {
                        total += *resource;
                    }
                }
            }
            return total;
        }
    };
    
    Module::Module(GameStateModule &gsm, History::HistoryModule & hist) : 
        gsm(gsm), hist(hist)
    {
        terrain = 0;
        gsm.add_handler(this);
    }

    bool Module::find_scouting_location(Loc &output) 
    {
        Vector<TerrainBase::Loc> border;
        FindBorderLocations(gsm.get_game().get_map(),
                            gsm.get_game().get_tile_points(),
                            border);
        if(border.empty())
        {
            return false;
        }
        else
        {
            size_t index = rand() % border.size();
            output = border[index];
            return true;
        }
    }

    void Module::find_resource_clusters(const std::string & bp_name,
                                        ClusterVec & results)
    {
        std::vector<const History::Record*> resources;
        std::vector<ResourceCluster*> clusters;

        hist.queryActiveRecords( Watcher::TypeIs(bp_name), resources );
        FORALL( resources, m )
        {
            const History::Record * resource = (*m);
            sint4 bestCluster = -1;
            real8 bestDistance = 128.0;
            for(sint4 i=0; i<(sint4)clusters.size(); i++)
            {
                real8 distance = clusters[i]->minDistance( (*m) );
                if(distance < bestDistance)
                {
                    bestCluster = i;
                    bestDistance = distance;
                }
            }
            if(bestCluster == -1)
            {
                clusters.push_back(new ResourceCluster(resource));
            }
            else
            {
                clusters[bestCluster]->addResource(resource);
            }
        }

        FORALL( clusters, c )
        {
            ResourceCluster * cluster = (*c);

            ResourceClusterInfo info;
            info.location = cluster->getCentroid();
            info.patchCount = cluster->countPatches();
            info.estimatedWorth = cluster->estimateWorth();
            results.push_back(info);

            delete cluster;
        }
        clusters.clear();
    }

    void Module::update_timeouts() 
    {
        //terrain->update_timeouts();
    }

    bool Module::handle_event(const Event &event) 
    {
        if (event.get_who() != GameStateModule::VIEW_MSG && 
            event.get_who() != GameStateModule::FROM)
            return false;
        
        if (!terrain)
        {
            terrain = new TerrainRegion(0, 0, 
                                        gsm.get_game().get_map().get_width(), 
                                        gsm.get_game().get_map().get_height());
        }
        
        const GameChanges& changes = gsm.get_changes();
        FORALL(changes.new_tile_indexes, i) 
        {
            terrain->add_tile(Loc(gsm.get_game().get_map().ind2x(*i), 
                                  gsm.get_game().get_map().ind2y(*i)));
        }

        return true;
    }

    Module::~Module() 
    {
        if (terrain) delete terrain;
        gsm.remove_handler(this);
    }
}
