/* 
 * TerrainAnalysis
 * Various Utilities to analysis placement of paths, units and buildings
 * over the terrain.  Built for Game 3 ORTS Comp 2007.
 *
 * Adam Elliot
*/
#ifndef TERRAINANALYSIS_H
#define TERRAINANALYSIS_H

#include <memory>

#include "Global.H"
#include "GameObj.H"
#include "SimpleTerrain.H"
#include "TerrainBase.H"
#include "EventHandler.H"
#include "Game.H"
#include "GameStateModule.H"
#include "GameChanges.H"
#include "History.H"

using namespace std;

#define SCOUT_TIMEOUT 8 /// Constant used to have units choose different tiles to scout

typedef TerrainBase::Loc Loc;

namespace TerrainAnalysis 
{
    /// Generates a vector of locations which are unknown but bordering
    /// a known tile.
    void FindBorderLocations(const Map<GameTile> & map, sint4 ppt,
                             Vector<TerrainBase::Loc> & results);

    class TerrainRegion 
    {
        ///  The size of the current region
        sint4 x, y;
        sint4 width, height;
        /// The quantity of unexplored tiles
        sint4 unexplored;
        /// How many times the region should be subdivided.
        sint4 submaps;

        /// The four subregions that this region is broken down into
        TerrainRegion *tl, *tr, *bl, *br;
    public:

        /**
         * Constructor for a region of terrain used to store information about
         * the currently explored terrain.
         * @param x The position of this grid in tiles from the left
         * @param y The position of this grid in tiles from the top
         * @param height The height of this subdivision. All subdivsions are
         * the same size on a level.
         * @param width The width of this subdivision.
         * @param submaps How many depths to be subdivided.
         */
        TerrainRegion(sint4 x, sint4 y, sint4 width, sint4 height, sint4 submaps = 3);

        /**
         * Indicateds that a tile has been viewed at a loc and removes it from
         * that unknown area.
         * @param loc The locatation of the tile in tile coordinates.
         */
        void add_tile(const Loc &loc);

        /**
         * Searches the map and it's submaps recusively until it finds the best
         * best area to be searched. This is the area with the highest quantity
         * of unexplored terrain. I've used a greedy algorithm here so their
         * may be cases where this isn't optimal, but should always give a
         * resonable result.
         * @param loc Returns the tile in tile coordinates that is the area to
         * explore.
         */
        void find_largest_unexplored_area(Loc &loc);

        /**
         * Returns how much area of this region is unexplored.
         */
        sint4 get_unexplored() { return unexplored; }

        /**
         * Clean up subregions.
         */
        ~TerrainRegion();
    };

        
    struct ResourceClusterInfo
    {
        TerrainBase::Loc location;
        sint4 patchCount;
        sint4 estimatedWorth;
    };

    class Module : public EventHandler 
    {
    public:
        typedef Vector<ResourceClusterInfo> ClusterVec;
    private:
        GameStateModule &gsm;
        History::HistoryModule &hist;
        /// Base region that contains the whole map.
        TerrainRegion *terrain;
    public:
        Module(GameStateModule &gsm,
               History::HistoryModule &hist);
        ~Module();

        bool handle_event(const Event &event);

        void update_timeouts();       
        bool find_scouting_location(Loc &output);
        void find_resource_clusters(const std::string & bp_name,
                                    ClusterVec & results);
    };
}

#endif // TERRAINANALYSIS_H
