#include "Watcher.H"

#include "GameStateModule.H"
#include "GameChanges.H"

#include <iostream>



namespace Watcher
{
    const sint4 WatcherModule::FROM      = EventFactory::new_who(); 
    const sint4 WatcherModule::CONDITION = EventFactory::new_what();
    const sint4 WatcherModule::CREATION  = EventFactory::new_what();
    const sint4 WatcherModule::DEATH     = EventFactory::new_what();

    // WATCH STRUCTURE

    WatcherModule::ConditionWatch::ConditionWatch() {}

    WatcherModule::ConditionWatch::ConditionWatch(const UnitPredPtr & query, 
                                const SetPredPtr & condition, 
                                EventHandler *handler,
                                uint4 period,
                                bool autoDelete) :
        query(query), condition(condition), handler(handler),
        ticks(0), period(period), autoDelete(autoDelete) {}

    WatcherModule::ConditionWatch & WatcherModule::ConditionWatch::operator = (
        const WatcherModule::ConditionWatch & o)
    {
        query      = o.query;
        condition  = o.condition;
        handler    = o.handler;
        ticks      = o.ticks;
        period     = o.period;
        autoDelete = o.autoDelete;
        return *this;
    }

    // CREATEWATCH STRUCTURE

    WatcherModule::CreationWatch::CreationWatch() {}

    WatcherModule::CreationWatch::CreationWatch(const UnitPredPtr & query, 
                                            EventHandler * handler,
                                            uint4 priority,
                                            bool autoDelete) :
        query(query), handler(handler), 
        priority(priority), autoDelete(autoDelete) {}

    WatcherModule::CreationWatch & WatcherModule::CreationWatch::operator = (
        const WatcherModule::CreationWatch & o)
    {
        query      = o.query;
        handler    = o.handler;
        priority   = o.priority;
        autoDelete = o.autoDelete;
        return *this;
    }

    // WATCHER MODULE

    WatcherModule::WatcherModule(GameStateModule& gsm) : 
        gsm(gsm), nextID(1)
    {
        gsm.add_handler(this);
    }

    WatcherModule::~WatcherModule()
    {

    }

    void WatcherModule::handleNew()
    {        
        std::map<WatchID, CreationEvent> mapCreate;

        // Iterate through all new units
        const GameChanges& gc = gsm.get_changes();
        Vector<ScriptObj::WPtr>::const_iterator it;
        for(it = gc.new_objs.begin(); it != gc.new_objs.end(); it++)
        {
            ScriptObj * sobj = (*it);
            GameObj * obj = dynamic_cast<GameObj*>(sobj);
            if( obj != NULL )
            {
                objects.insert( obj );

                // Find the highest-priority creation watch
                WatchID best; sint4 highest = -1;
                std::map<WatchID, CreationWatch>::const_iterator cit;
                for(cit = mapCreationWatches.begin(); 
                    cit != mapCreationWatches.end(); cit++)
                {
                    const CreationWatch & cwatch = cit->second;
                    if( cwatch.query->evaluate( obj ) )
                    {
                        sint4 priority = static_cast<sint4>(cwatch.priority);
                        if( priority > highest )
                        {
                            highest = priority;
                            best    = cit->first;
                        }
                    }
                }

                // If a handler is interested in this unit
                if( highest > -1 )
                {
                    // Add the unit to this creation event
                    mapCreate[best].setID(best);
                    mapCreate[best].getUnits().insert( obj );
                }
            }
        }

        // Dispatch any creation events
        std::map<WatchID, CreationEvent>::iterator it2, it_rem;
        for(it2 = mapCreate.begin(); it2 != mapCreate.end();)
        {
            it_rem = it2++;
            EventHandler* handler = mapCreationWatches[it_rem->first].handler;

            if(mapCreationWatches[it_rem->first].autoDelete)
            {
                mapCreationWatches.erase(it_rem->first);
            }

            handler->handle_event(it_rem->second);
        }
    }

    void WatcherModule::handleDeaths()
    {
        std::map<EventHandler*, DeathEvent> mapDeath;

        const GameChanges& gc = gsm.get_changes();

        // Iterate through all dead or vanished units
        Vector<ScriptObj::WPtr> gone_objs(gc.vanished_objs);
        copy(gc.dead_objs.begin(), gc.dead_objs.end(), 
             std::back_inserter(gone_objs));        

        Vector<ScriptObj::WPtr>::const_iterator it;
        for(it = gone_objs.begin(); it != gone_objs.end(); it++)
        {
            ScriptObj * sobj = (*it);
            GameObj * obj = dynamic_cast<GameObj*>(sobj);
            if( obj != NULL )
            {
                // If we were watching this unit
                std::multimap<GameObj*, EventHandler*>::iterator 
                    it_l, it_u, it;
                it_l = mapDeathHandlers.lower_bound( obj );
                it_u = mapDeathHandlers.upper_bound( obj );
                for(it = it_l; it != it_u; it++)
                {
                    // Stop watching the unit
                    EventHandler* owner = it->second;

                    // And record its death
                    mapDeath[owner].getUnits().insert( obj );
                }
                mapDeathHandlers.erase( obj );

                // Forget about the unit
                objects.erase( obj );
            }            
        }

        // Dispatch death notifications
        std::map<EventHandler*, DeathEvent>::iterator dIt;
        for(dIt = mapDeath.begin(); dIt != mapDeath.end(); dIt++)
        {
            EventHandler* owner = dIt->first;
            DeathEvent& event = dIt->second;
            owner->handle_event(event);
        }
    }

    void WatcherModule::checkWatches()
    {
        // Iterate through all of our triggers
        std::map<WatchID, ConditionWatch>::iterator it, it2;
        for(it = mapWatches.begin(); it != mapWatches.end();)
        {
            // Examine a specific trigger
            it2 = it++;
            ConditionWatch & trigger = it2->second;

            // If we haven't waited long enough
            trigger.ticks++;
            if(trigger.ticks < trigger.period)
            {
                // Wait some more
                continue;
            }
            trigger.ticks = 0;

            // Create an event with the current trigger ID
            ConditionEvent event(it2->first);

            // Perform the query
            queryUnits(trigger.query, event.getUnits());

            // If the results pass the predicate
            if(trigger.condition->evaluate(event.getUnits()))
            {
                // Determine the event handler
                EventHandler* handler = trigger.handler;

                // If we are set to autodelete
                if(trigger.autoDelete)
                {
                    // Delete this trigger
                    mapWatches.erase(it2);
                }

                // Dispatch the event
                handler->handle_event(event);
            }
        }
    }

    bool WatcherModule::handle_event(const Event &event)
    {
        if( event.get_who() != GameStateModule::FROM )
        {
            return false;
        }

        if( event.get_what() != GameStateModule::VIEW_MSG )
        {
            return false;
        }

        // Handle new objects
        handleNew();

        // Handle dead objects
        handleDeaths();

        // Check for new notifications
        checkWatches();

        return true;
    }

    size_t WatcherModule::queryUnits(const UnitPredPtr & query, 
                                     std::set<GameObj*> & result) const
    {
        size_t count = 0;
        std::set<GameObj*>::const_iterator it;
        for(it = objects.begin(); it != objects.end(); it++)
        {
            if( query->evaluate( (*it) ) )
            {
                result.insert( (*it) );
                count++;
            }
        }
        return count;
    }

    WatchID WatcherModule::notifyOnCondition(EventHandler *handler,
                                             const UnitPredPtr & query,
                                             const SetPredPtr & condition,
                                             uint4 period,
                                             bool autoDelete)
    {
        // TODO: Robust selection of nextID

        mapWatches[nextID] = ConditionWatch(query, condition, handler, 
                                            period, autoDelete);
        return nextID++;
    }

    WatchID WatcherModule::notifyOnCreation(EventHandler *handler,
                                            const UnitPredPtr & query,
                                            uint4 priority,
                                            bool autoDelete)
    {
        // TODO: Robust selection of nextID

        mapCreationWatches[nextID] = CreationWatch(query, handler, 
                                                   priority, autoDelete);

        return nextID++;
    }

    void WatcherModule::notifyOnDeath(EventHandler *handler,
                                      GameObj * unit)
    {
        mapDeathHandlers.insert(std::make_pair(unit, handler));
    }
    
    void WatcherModule::cancelNotify(WatchID id)
    {
        // If a regular watch, delete it
        std::map<WatchID, ConditionWatch>::iterator it;
        it = mapWatches.find(id);
        if(it != mapWatches.end())
        {
            mapWatches.erase(it);
        }

        // If a creation watch, delete it
        std::map<WatchID, CreationWatch>::iterator it2;
        it2 = mapCreationWatches.find(id);
        if(it2 != mapCreationWatches.end())
        {
            mapCreationWatches.erase(it2);
        }
    }

    void WatcherModule::cancelNotifyOnDeath(EventHandler * handler,
                                            GameObj * unit)
    {
        std::multimap<GameObj*, EventHandler*>::iterator it_l, it_u, it;
        it_l = mapDeathHandlers.lower_bound( unit );
        it_u = mapDeathHandlers.upper_bound( unit );
        for(it = it_l; it != it_u; it++)
        {
            if(it->second == handler)
            {
                mapDeathHandlers.erase(it);
                return;
            }
        }
    }

    // TRIGGER EVENT

    ConditionEvent::ConditionEvent(WatchID id) : 
        Event(WatcherModule::FROM, WatcherModule::CONDITION), id(id)
    {}

    ConditionEvent::~ConditionEvent() {}

    WatchID ConditionEvent::getID() const
    {
        return id;
    }

    std::set<GameObj*> & ConditionEvent::getUnits()
    {
        return units;
    }

    const std::set<GameObj*> & ConditionEvent::getUnits() const
    {
        return units;
    }  

    // CREATE EVENT

    CreationEvent::CreationEvent() : 
        Event(WatcherModule::FROM, WatcherModule::CREATION), id(id)
    {}

    CreationEvent::~CreationEvent() {}

    void CreationEvent::setID(WatchID id)
    {
        this->id = id;
    }

    WatchID CreationEvent::getID() const
    {
        return id;
    }

    std::set<GameObj*> & CreationEvent::getUnits()
    {
        return units;
    }

    const std::set<GameObj*> & CreationEvent::getUnits() const
    {
        return units;
    }

    // DEATH EVENT

    DeathEvent::DeathEvent() : 
        Event(WatcherModule::FROM, WatcherModule::DEATH)
    {}

    DeathEvent::~DeathEvent() {}

    std::set<GameObj*> & DeathEvent::getUnits()
    {
        return units;
    }

    const std::set<GameObj*> & DeathEvent::getUnits() const
    {
        return units;
    }  
}
