#include "WatcherAux.H"

#include "GameObj.H"
#include "Blueprints.H"

namespace Watcher
{
    const std::string WORKER         ("worker");
    const std::string MARINE         ("marine");
    const std::string TANK           ("tank");

    const std::string CONTROL_CENTER ("controlCenter");
    const std::string BARRACKS       ("barracks");
    const std::string FACTORY        ("factory");


    // Utility class to access a value
    class UnitValue
    {
    public:
        UnitValue() {}
        virtual ~UnitValue() {}
        
        virtual bool evaluate(const GameObj *object, sint4 &result) const = 0;
    };
    typedef boost::shared_ptr<UnitValue> UnitValuePtr;

    class CachedVariable : public UnitValue
    {
        std::string                      var_name;
        mutable const GameObjBlueprint * last_bp;
        mutable sint4                    index;
    public:
        CachedVariable(const std::string &var_name) :
            var_name(var_name), last_bp(NULL) {}
        CachedVariable & operator = (const CachedVariable & o)
        {
            var_name = o.var_name;
            last_bp  = o.last_bp;
            index    = o.index;
            return *this;
        }

        bool evaluate(const GameObj * object, sint4 & result) const
        {
            const GameObjBlueprint * bp = object->get_bp();
            if( bp != last_bp )
            {
                // TODO: Fix constness of many
                // methods of GameObjectBlueprint
                GameObjBlueprint * nonconst_bp =
                    const_cast<GameObjBlueprint*>(bp);
                index   = nonconst_bp->attr_index(var_name);
                last_bp = bp;
            }

            if(index < 0) 
            {
                return false;
            }

            const stype * ptr = object->get_val_ptr(index);
            result = ptr->to_int();
            return true;
        }
    };

    // Access a sub-component's variable
    class SubVariable : public UnitValue
    {
        std::string                      comp_name;
        mutable const GameObjBlueprint * last_bp;
        mutable sint4                    index;
        CachedVariable                   var;
    public:
        SubVariable(const std::string &comp_name, const std::string &var_name):
            comp_name(comp_name), last_bp(NULL), var(var_name) {}
        SubVariable & operator = (const SubVariable & o)
        {
            comp_name = o.comp_name;
            last_bp   = o.last_bp;
            index     = o.index;
            var       = o.var;
            return *this;
        }

        bool evaluate(const GameObj * object, sint4 & result) const
        {
            const GameObjBlueprint * bp = object->get_bp();
            if( bp != last_bp )
            {
                index   = object->component_index(comp_name);
                last_bp = bp;
            }

            if(index < 0) 
            {
                return false;
            }

            const ScriptObj * s_comp = object->component(index);
            const GameObj * comp = dynamic_cast<const GameObj*>(s_comp);
            if(comp == NULL) return false;
            return var.evaluate(comp, result);
        }
    };

    // Access the percentile ratio of two variables
    class PercentValue : public UnitValue
    {
        CachedVariable var_num;
        CachedVariable var_denom;
    public:
        PercentValue(const std::string & prop_num,
                      const std::string & prop_denom) :
            var_num(prop_num), var_denom(prop_denom) {}
        PercentValue & operator = (const PercentValue & o)
        {
            var_num   = o.var_num;
            var_denom = o.var_denom;
            return *this;
        }
        
        bool evaluate(const GameObj * unit, sint4 & result) const
        {
            sint4 num, denom;
            if(!var_num.evaluate(unit, num)) return false;
            if(!var_denom.evaluate(unit, denom)) return false;
            if(denom == 0) return false;
            result = 100 * num / denom;
            return true;
        }
    };

    class BattleStrength : public UnitValue
    {
        CachedVariable varHealth;
        SubVariable    varMinDmg;
        SubVariable    varMaxDmg;
        SubVariable    varCooldown;
    public:
        BattleStrength() : 
            varHealth("hp"), 
            varMinDmg("weapon", "min_damage"), 
            varMaxDmg("weapon", "max_damage"), 
            varCooldown("weapon", "cooldown")
        {}
        ~BattleStrength() {}

        bool evaluate(const GameObj * unit, sint4 & result) const
        {
            sint4 hp, min, max, cool;
            if( !varHealth.evaluate(unit, hp) ) return false;
            if( !varMinDmg.evaluate(unit, min) ) return false;
            if( !varMaxDmg.evaluate(unit, max) ) return false;
            if( !varCooldown.evaluate(unit, cool) ) return false;
            if( cool == 0 ) return false;
            result = hp*(min+max)/(cool*2);
            return true;
        }
    };

    // Utility functions to perform comparison
    typedef bool (*CompareFunc)(sint4 a, sint4 b);

    bool cmp_lt(sint4 a, sint4 b)  { return a < b; }
    bool cmp_gt(sint4 a, sint4 b)  { return a > b; }
    bool cmp_lte(sint4 a, sint4 b) { return a <= b; }
    bool cmp_gte(sint4 a, sint4 b) { return a >= b; }
    bool cmp_eq(sint4 a, sint4 b)  { return a == b; }
    bool cmp_neq(sint4 a, sint4 b) { return a != b; }

    CompareFunc GetCompareFunc(Comparator comp)
    {
        switch(comp)
        {
            case LESS_THAN: return cmp_lt;
            case GREATER_THAN: return cmp_gt;
            case LESS_THAN_EQUAL: return cmp_lte;
            case GREATER_THAN_EQUAL: return cmp_gte;
            case EQUAL_TO: return cmp_eq;
            default: return cmp_neq;
        }
    }

    // CONVENIENCE METHODS
    class NotPredicate : public UnitPred
    {
        UnitPredPtr p0;
    public:
        NotPredicate(const UnitPredPtr & p0) : p0(p0) {}
        ~NotPredicate() {}

        bool evaluate(const GameObj * unit) const
        {
            return !(p0->evaluate(unit));
        }
    };

    class AndPredicate : public UnitPred
    {
        UnitPredPtr p0;
        UnitPredPtr p1;
    public:
        AndPredicate(const UnitPredPtr & p0, const UnitPredPtr & p1) : 
            p0(p0), p1(p1) {}
        ~AndPredicate() {}

        bool evaluate(const GameObj * unit) const
        {
            if(p0->evaluate(unit) == false) return false;
            return p1->evaluate(unit);
        }
    };

    class OrPredicate : public UnitPred
    {
        UnitPredPtr p0;
        UnitPredPtr p1;
    public:
        OrPredicate(const UnitPredPtr & p0, const UnitPredPtr & p1) : 
            p0(p0), p1(p1) {}
        ~OrPredicate() {}

        bool evaluate(const GameObj * unit) const
        {
            if(p0->evaluate(unit) == true) return true;
            return p1->evaluate(unit);
        }
    };

    // UNIT PREDICATE CLASSES

    class UnitBPPredicate : public UnitPred
    {
        std::string bp_name;
        mutable const GameObjBlueprint * bp;
    public:
        UnitBPPredicate(const std::string & bp_name) :
            bp_name(bp_name), bp(NULL) {}
        ~UnitBPPredicate() {}

        bool evaluate(const GameObj * unit) const
        {
            return unit->bp_name() == bp_name;
        }   
    };

    class UnitValuePredicate : public UnitPred
    {
        UnitValuePtr value;
        CompareFunc  compare;
        sint4        ref_value;
    public:
        UnitValuePredicate(const UnitValuePtr & value, 
                           CompareFunc compare, sint4 ref_value) : 
            value(value), compare(compare), ref_value(ref_value) {}
        ~UnitValuePredicate() {}

        bool evaluate(const GameObj * unit) const
        {
            sint4 unit_value;
            if( !value->evaluate(unit, unit_value) ) return false;
            return compare(unit_value, ref_value);
        }
    };

    class UnitSODPredicate : public UnitPred
    {
        sintptr * ServerObjData::* data_ptr;
        CompareFunc                compare;
        sint4                      ref_value;
    public:
        UnitSODPredicate(sintptr * ServerObjData::*data_ptr,
                         CompareFunc compare, sint4 ref_value) :
            data_ptr(data_ptr), compare(compare), ref_value(ref_value) {}
        ~UnitSODPredicate() {}

        bool evaluate(const GameObj * unit) const
        {
            sintptr * data = unit->sod.*data_ptr;
            return compare((*data), ref_value);
        }
    };

    class UnitComparePredicate : public UnitPred
    {
        UnitValuePtr left;
        UnitValuePtr right;
        CompareFunc  compare;
    public:
        UnitComparePredicate(const UnitValuePtr & left, 
                             const UnitValuePtr & right,
                             CompareFunc compare) : 
            left(left), right(right), compare(compare) {}
        ~UnitComparePredicate() {}

        bool evaluate(const GameObj * unit) const
        {
            sint4 left_value, right_value;
            if( !left->evaluate(unit, left_value) ) return false;
            if( !right->evaluate(unit, right_value) ) return false;
            return compare(left_value, right_value);
        }
    };

    class InCirclePred : public UnitPred
    {
        sint4 x;
        sint4 y;
        sint4 r2;
    public:
        InCirclePred(sint4 x, sint4 y, sint4 r) : x(x), y(y), r2(r*r) {}
        ~InCirclePred() {}

        bool evaluate(const GameObj * unit) const
        {
	    if( unit->get_shape() == Object::SHAPE_UNDEF ) return false;
            sint4 px, py;
            unit->get_center(px, py);
            sint4 dx = px - x;
            sint4 dy = py - y;
            return ((dx*dx + dy*dy) <= r2);
        }
    };

    class InRectanglePred : public UnitPred
    {
        sint4 x0;
        sint4 y0;
        sint4 x1;
        sint4 y1;
    public:
        InRectanglePred(sint4 x0, sint4 y0, sint4 x1, sint4 y1) : 
            x0(x0), y0(y0), x1(x1), y1(y1) {}
        ~InRectanglePred() {}

        bool evaluate(const GameObj * unit) const
        {
	    if( unit->get_shape() == Object::SHAPE_UNDEF ) return false;
            sint4 px, py;
            unit->get_center(px, py);
            if(px < x0) return false;
            if(py < y0) return false;
            if(px > x1) return false;
            if(py > y1) return false;
            return true;
        }
    };

    UnitPredPtr PropertyIs(const UnitValuePtr & value,
                           CompareFunc compare, sint4 ref_value)
    {
        return UnitPredPtr(new UnitValuePredicate(value, compare, ref_value));
    }

    UnitPredPtr PropertyIs(const UnitValuePtr & value,
                           Comparator comp, sint4 ref_value)
    {
        return PropertyIs(value, GetCompareFunc(comp), ref_value);
    }

    // Convenience functions for making unit predicates

    UnitPredPtr Not(const UnitPredPtr & p0)
    {
        return UnitPredPtr(new NotPredicate(p0));
    }

    UnitPredPtr And(const UnitPredPtr & p0, const UnitPredPtr & p1)
    {
        return UnitPredPtr(new AndPredicate(p0, p1));
    }

    UnitPredPtr And(const UnitPredPtr & p0, const UnitPredPtr & p1, 
                    const UnitPredPtr & p2)
    {
        return And(p0, And(p1, p2));
    }

    UnitPredPtr And(const UnitPredPtr & p0, const UnitPredPtr & p1, 
                    const UnitPredPtr & p2, const UnitPredPtr & p3)
    {
        return And(p0, And(p1, And(p2, p3)));
    }

    UnitPredPtr Or(const UnitPredPtr & p0, const UnitPredPtr & p1)
    {
        return UnitPredPtr(new OrPredicate(p0, p1));
    }

    UnitPredPtr Or(const UnitPredPtr & p0, const UnitPredPtr & p1, 
                    const UnitPredPtr & p2)
    {
        return Or(p0, Or(p1, p2));
    }

    UnitPredPtr Or(const UnitPredPtr & p0, const UnitPredPtr & p1, 
                    const UnitPredPtr & p2, const UnitPredPtr & p3)
    {
        return Or(p0, Or(p1, Or(p2, p3)));
    }

    UnitPredPtr TypeIs(const std::string & type)
    {
        return UnitPredPtr(new UnitBPPredicate(type));
    }

    UnitPredPtr IsBuilding()
    {
        return Or(TypeIs(CONTROL_CENTER), TypeIs(BARRACKS), TypeIs(FACTORY));
    }

    UnitPredPtr IsWorker()
    {
        return TypeIs(WORKER);
    }

    UnitPredPtr IsCombat()
    {
        return Or(TypeIs(MARINE), TypeIs(TANK));
    }

    UnitPredPtr BuildingDone()
    {
        return UnitPredPtr(new UnitComparePredicate(
             UnitValuePtr(new CachedVariable("amount_built")),
             UnitValuePtr(new CachedVariable("build_time")),
             GetCompareFunc(GREATER_THAN_EQUAL)));
    }

    UnitPredPtr OwnerIs(sint4 owner)
    {
        return UnitPredPtr(new UnitSODPredicate(&ServerObjData::owner, 
                                                GetCompareFunc(EQUAL_TO), 
                                                owner));
    }

    UnitPredPtr HealthIs(Comparator comp, sint4 value)
    {
        return PropertyIs(UnitValuePtr(new CachedVariable("hp")), 
                          comp, value);
    } 

    UnitPredPtr HealthPctIs(Comparator comp, sint4 value)
    {
        return PropertyIs(UnitValuePtr(new PercentValue("hp", "max_hp")), 
                          comp, value);
    } 

    UnitPredPtr StrengthIs(Comparator comp, sint4 value)
    {
        return PropertyIs(UnitValuePtr(new BattleStrength()), comp, value);
    } 

    UnitPredPtr InCircle(sint4 x, sint4 y, sint4 r)
    {
        return UnitPredPtr(new InCirclePred(x,y,r));
    }

    UnitPredPtr InRectangle(sint4 x0, sint4 y0, sint4 x1, sint4 y1)
    {
        return UnitPredPtr(new InRectanglePred(x0,y0,x1,y1));
    }

    sint4 StrengthOf(const GameObj * unit)
    {
        static BattleStrength strength;
        sint4 result;
        return strength.evaluate(unit, result) ? result : 0;
    }

    // SET PREDICATES

    template<class ACCESSOR, class COMP>
    class SetValuePredicate : public SetPred
    {
        ACCESSOR accessor;
        sint4    ref_value;
    public:
        SetValuePredicate(const ACCESSOR & accessor, sint4 ref_value) : 
            accessor(accessor), ref_value(ref_value) {}
        ~SetValuePredicate() {}

        bool evaluate(const std::set<GameObj*> & set) const
        {
            COMP comp;
            sint4 set_value = accessor(set);
            return comp(set_value, ref_value);
        }
    };  

    template<typename ACCESSOR, typename COMP>
    SetPredPtr SetPropertyIs(const ACCESSOR & accessor, sint4 value)
    {
        return SetPredPtr(
            new SetValuePredicate<ACCESSOR, COMP>(accessor, value));
    }

    template<typename ACCESSOR>
    SetPredPtr SetPropertyIs(const ACCESSOR & accessor,
                             Comparator comp, sint4 value)
    {
        switch(comp)
        {
           case LESS_THAN:
               return SetPropertyIs<ACCESSOR, std::less<sint4> >(
                   accessor, value);
           case GREATER_THAN:
               return SetPropertyIs<ACCESSOR, std::greater<sint4> >(
                   accessor, value);
           case LESS_THAN_EQUAL:
               return SetPropertyIs<ACCESSOR, std::less_equal<sint4> >(
                   accessor, value);
           case GREATER_THAN_EQUAL:
               return SetPropertyIs<ACCESSOR, std::greater_equal<sint4> >(
                   accessor, value);
           case EQUAL_TO:
               return SetPropertyIs<ACCESSOR, std::equal_to<sint4> >(
                   accessor, value);
           default:
               return SetPropertyIs<ACCESSOR, std::not_equal_to<sint4> >(
                   accessor, value);
        }
    }

    struct SetCount
    {
        sint4 operator()(const std::set<GameObj*> & set) const
        {
            return set.size();
        }
    };

    struct SetStrength
    {
        BattleStrength strength;

        sint4 operator()(const std::set<GameObj*> & set) const
        {
            sint4 total = 0;
            std::set<GameObj*>::const_iterator it;
            for(it = set.begin(); it != set.end(); it++)
            {
                sint4 value;
                if( strength.evaluate((*it), value) )
                {
                    total += value;
                }
            }
            return total;
        }
    };

    SetPredPtr TotalCountIs(Comparator comp, sint4 value)
    {
        return SetPropertyIs(SetCount(), comp, value);
    }

    SetPredPtr TotalStrengthIs(Comparator comp, sint4 value)
    {
        return SetPropertyIs(SetStrength(), comp, value);
    }
}
